#!/usr/bin/env python3

"""Meta data about tables and ids for self serve deletion."""

import logging
import re
from collections import defaultdict
from dataclasses import dataclass
from functools import partial
from itertools import chain
from multiprocessing.pool import ThreadPool
from typing import Dict, List

import requests
from google.cloud import bigquery
from google.cloud.exceptions import NotFound

from ..util.bigquery_id import qualified_table_id

SHARED_PROD = "moz-fx-data-shared-prod"
GLEAN_SCHEMA_ID = "glean_ping_1"
GLEAN_APP_LISTINGS_URL = "https://probeinfo.telemetry.mozilla.org/v2/glean/app-listings"


@dataclass(frozen=True)
class DeleteSource:
    """Data class for deletion request source."""

    table: str
    field: str
    project: str = SHARED_PROD
    conditions: tuple[str, ...] = ()

    def __post_init__(self):
        """Validate the table string."""
        if len(self.table.split(".")) != 2:
            raise ValueError(
                "DeleteSource table must be in the 'dataset.table' format."
            )

    @property
    def table_id(self):
        """Table Id."""
        return self.table.split(".", 1)[-1]

    @property
    def dataset_id(self):
        """Dataset Id."""
        return self.table.split(".", 1)[0]


@dataclass(frozen=True)
class DeleteTarget:
    """Data class for deletion request target.

    Rows will be removed using either one DELETE statement for the whole table,
    or one DELETE statement per partition if the table is larger than some
    configurable threshold.
    """

    table: str
    field: str | tuple[str, ...]
    project: str = SHARED_PROD

    def __post_init__(self):
        """Validate the table string."""
        if len(self.table.split(".")) != 2:
            raise ValueError(
                "DeleteTarget table must be in the 'dataset.table' format."
            )

    @property
    def table_id(self):
        """Table Id."""
        return self.table.partition(".")[2]

    @property
    def dataset_id(self):
        """Dataset Id."""
        return self.table.partition(".")[0]

    @property
    def fields(self) -> tuple[str, ...]:
        """Fields."""
        if isinstance(self.field, tuple):
            return self.field
        return (self.field,)


DeleteIndex = dict[DeleteTarget, DeleteSource | tuple[DeleteSource, ...]]

CLIENT_ID = "client_id"
GLEAN_CLIENT_ID = "client_info.client_id"
IMPRESSION_ID = "impression_id"
USER_ID = "user_id"
POCKET_ID = "pocket_id"
SHIELD_ID = "shield_id"
PIONEER_ID = "pioneer_id"
RALLY_ID = "metrics.uuid.rally_id"
RALLY_ID_TOP_LEVEL = "rally_id"
ID = "id"
FXA_USER_ID = "jsonPayload.fields.user_id"
# these must be in the same order as SYNC_SOURCES
SYNC_IDS = ("SUBSTR(payload.device_id, 0, 32)", "payload.uid")
CONTEXT_ID = "context_id"
USER_CHARACTERISTICS_ID = "metrics.uuid.characteristics_client_identifier"

DESKTOP_SRC = DeleteSource(
    table="telemetry_stable.deletion_request_v4", field=CLIENT_ID
)
DESKTOP_GLEAN_SRC = DeleteSource(
    table="firefox_desktop_stable.deletion_request_v1", field=GLEAN_CLIENT_ID
)
IMPRESSION_SRC = DeleteSource(
    table="telemetry_stable.deletion_request_v4",
    field="payload.scalars.parent.deletion_request_impression_id",
)
CONTEXTUAL_SERVICES_SRC = DeleteSource(
    table="telemetry_stable.deletion_request_v4",
    field="payload.scalars.parent.deletion_request_context_id",
)
FXA_HMAC_SRC = DeleteSource(
    table="firefox_accounts.fxa_delete_events", field="hmac_user_id"
)
FXA_SRC = DeleteSource(table="firefox_accounts.fxa_delete_events", field=USER_ID)
FXA_UNHASHED_SRC = DeleteSource(
    table="firefox_accounts.fxa_delete_events", field="user_id_unhashed"
)
FXA_FRONTEND_GLEAN_SRC = DeleteSource(
    table="accounts_frontend_stable.deletion_request_v1", field=GLEAN_CLIENT_ID
)
REGRETS_SRC = DeleteSource(
    table="regrets_reporter_stable.regrets_reporter_update_v1",
    field="data_deletion_request.extension_installation_uuid",
    conditions=("data_deletion_request IS NOT NULL",),
)
# these must be in the same order as SYNC_IDS
SYNC_SOURCES = (
    DeleteSource(
        table="telemetry_stable.deletion_request_v4",
        field="payload.scalars.parent.deletion_request_sync_device_id",
    ),
    DeleteSource(
        table="firefox_accounts.fxa_delete_events",
        field="SUBSTR(hmac_user_id, 0, 32)",
    ),
)
LEGACY_MOBILE_SOURCES = tuple(
    DeleteSource(
        table=f"{product}_stable.deletion_request_v1",
        field="metrics.uuid.legacy_ids_client_id",
    )
    for product in (
        "org_mozilla_ios_fennec",
        "org_mozilla_ios_firefox",
        "org_mozilla_ios_firefoxbeta",
        "org_mozilla_tv_firefox",
        "mozilla_lockbox",
    )
)
USER_CHARACTERISTICS_SRC = DeleteSource(
    table="firefox_desktop_stable.deletion_request_v1",
    field=USER_CHARACTERISTICS_ID,
)
SOURCES = (
    [
        DESKTOP_SRC,
        IMPRESSION_SRC,
        CONTEXTUAL_SERVICES_SRC,
        FXA_HMAC_SRC,
        FXA_SRC,
    ]
    + list(SYNC_SOURCES)
    + list(LEGACY_MOBILE_SOURCES)
)

LEGACY_MOBILE_IDS = tuple(CLIENT_ID for _ in LEGACY_MOBILE_SOURCES)


client_id_target = partial(DeleteTarget, field=CLIENT_ID)
glean_target = partial(DeleteTarget, field=GLEAN_CLIENT_ID)
impression_id_target = partial(DeleteTarget, field=IMPRESSION_ID)
fxa_user_id_target = partial(DeleteTarget, field=FXA_USER_ID)
user_id_target = partial(DeleteTarget, field=USER_ID)
context_id_target = partial(DeleteTarget, field=CONTEXT_ID)

DELETE_TARGETS: DeleteIndex = {
    # Other
    client_id_target(table="search_derived.acer_cohort_v1"): DESKTOP_SRC,
    client_id_target(
        table="search_derived.mobile_search_clients_daily_v1"
    ): DESKTOP_SRC,
    client_id_target(
        table="search_derived.mobile_search_clients_daily_v2"
    ): DESKTOP_SRC,
    client_id_target(table="search_derived.search_clients_daily_v8"): DESKTOP_SRC,
    client_id_target(
        table="telemetry_derived.desktop_engagement_clients_v1"
    ): DESKTOP_SRC,
    client_id_target(table="search_derived.search_clients_last_seen_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.clients_daily_v6"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.clients_daily_joined_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.clients_last_seen_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.clients_last_seen_v2"): DESKTOP_SRC,
    client_id_target(
        table="telemetry_derived.clients_last_seen_joined_v1"
    ): DESKTOP_SRC,
    client_id_target(
        table="telemetry_derived.clients_profile_per_install_affected_v1"
    ): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.core_clients_daily_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.core_clients_last_seen_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.event_events_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.main_events_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.main_1pct_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.main_remainder_1pct_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.main_use_counter_1pct_v1"): DESKTOP_SRC,
    client_id_target(
        table="telemetry_derived.desktop_retention_clients_v1"
    ): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.block_autoplay_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.crash_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.downgrade_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.event_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.first_shutdown_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.first_shutdown_v5"): DESKTOP_SRC,
    client_id_target(
        table="telemetry_stable.first_shutdown_use_counter_v4"
    ): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.focus_event_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.frecency_update_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.health_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.heartbeat_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.main_v5"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.main_use_counter_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.modules_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.new_profile_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.saved_session_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.saved_session_v5"): DESKTOP_SRC,
    client_id_target(
        table="telemetry_stable.saved_session_use_counter_v4"
    ): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.shield_icq_v1_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.shield_study_addon_v3"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.shield_study_error_v3"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.shield_study_v3"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.testpilot_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.third_party_modules_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.untrusted_modules_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.update_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.voice_v4"): DESKTOP_SRC,
    DeleteTarget(
        table="telemetry_derived.mobile_engagement_clients_v1",
        field=(CLIENT_ID, CLIENT_ID),
    ): (
        DeleteSource(table="firefox_ios.deletion_request", field=GLEAN_CLIENT_ID),
        DeleteSource(table="fenix.deletion_request", field=GLEAN_CLIENT_ID),
    ),
    # activity stream
    DeleteTarget(
        table="messaging_system_stable.cfr_v1", field=(CLIENT_ID, IMPRESSION_ID)
    ): (DESKTOP_SRC, IMPRESSION_SRC),
    DeleteTarget(
        table="messaging_system_derived.cfr_users_daily_v1",
        field=(CLIENT_ID, IMPRESSION_ID),
    ): (DESKTOP_SRC, IMPRESSION_SRC),
    DeleteTarget(
        table="messaging_system_derived.cfr_users_last_seen_v1",
        field=(CLIENT_ID, IMPRESSION_ID),
    ): (DESKTOP_SRC, IMPRESSION_SRC),
    client_id_target(table="activity_stream_stable.events_v1"): DESKTOP_SRC,
    client_id_target(table="messaging_system_stable.onboarding_v1"): DESKTOP_SRC,
    client_id_target(table="messaging_system_stable.snippets_v1"): DESKTOP_SRC,
    client_id_target(table="activity_stream_stable.sessions_v1"): DESKTOP_SRC,
    client_id_target(
        table="messaging_system_derived.onboarding_users_daily_v1"
    ): DESKTOP_SRC,
    client_id_target(
        table="messaging_system_derived.onboarding_users_last_seen_v1"
    ): DESKTOP_SRC,
    client_id_target(
        table="messaging_system_derived.snippets_users_daily_v1"
    ): DESKTOP_SRC,
    client_id_target(
        table="messaging_system_derived.snippets_users_last_seen_v1"
    ): DESKTOP_SRC,
    impression_id_target(
        table="activity_stream_stable.impression_stats_v1"
    ): IMPRESSION_SRC,
    impression_id_target(table="activity_stream_stable.spoc_fills_v1"): IMPRESSION_SRC,
    impression_id_target(
        table="messaging_system_stable.undesired_events_v1"
    ): IMPRESSION_SRC,
    impression_id_target(
        table="messaging_system_stable.personalization_experiment_v1"
    ): IMPRESSION_SRC,
    # sync
    DeleteTarget(table="telemetry_stable.sync_v4", field=SYNC_IDS): SYNC_SOURCES,
    DeleteTarget(table="telemetry_stable.sync_v5", field=SYNC_IDS): SYNC_SOURCES,
    # fxa
    user_id_target(
        table="firefox_accounts_derived.fxa_amplitude_export_v1"
    ): FXA_HMAC_SRC,
    user_id_target(
        table="firefox_accounts_derived.fxa_amplitude_user_ids_v1"
    ): FXA_HMAC_SRC,
    fxa_user_id_target(
        table="firefox_accounts_derived.fxa_auth_bounce_events_v1"
    ): FXA_SRC,
    fxa_user_id_target(table="firefox_accounts_derived.fxa_auth_events_v1"): FXA_SRC,
    fxa_user_id_target(table="firefox_accounts_derived.fxa_content_events_v1"): FXA_SRC,
    fxa_user_id_target(
        table="firefox_accounts_derived.fxa_gcp_stderr_events_v1"
    ): FXA_SRC,
    fxa_user_id_target(
        table="firefox_accounts_derived.fxa_gcp_stdout_events_v1"
    ): FXA_SRC,
    user_id_target(
        table="firefox_accounts_derived.fxa_log_device_command_events_v1"
    ): FXA_HMAC_SRC,
    user_id_target(
        table="firefox_accounts_derived.fxa_log_device_command_events_v2"
    ): FXA_HMAC_SRC,
    fxa_user_id_target(table="firefox_accounts_derived.fxa_oauth_events_v1"): FXA_SRC,
    fxa_user_id_target(table="firefox_accounts_derived.fxa_stdout_events_v1"): FXA_SRC,
    user_id_target(table="firefox_accounts_derived.fxa_users_daily_v1"): FXA_SRC,
    user_id_target(table="firefox_accounts_derived.fxa_users_daily_v2"): FXA_SRC,
    user_id_target(table="firefox_accounts_derived.fxa_users_first_seen_v1"): FXA_SRC,
    user_id_target(table="firefox_accounts_derived.fxa_users_first_seen_v2"): FXA_SRC,
    user_id_target(table="firefox_accounts_derived.fxa_users_last_seen_v1"): FXA_SRC,
    user_id_target(table="firefox_accounts_derived.fxa_users_last_seen_v2"): FXA_SRC,
    user_id_target(
        table="firefox_accounts_derived.fxa_users_services_daily_v1"
    ): FXA_SRC,
    user_id_target(
        table="firefox_accounts_derived.fxa_users_services_daily_v2"
    ): FXA_SRC,
    user_id_target(
        table="firefox_accounts_derived.fxa_users_services_first_seen_v1"
    ): FXA_SRC,
    user_id_target(
        table="firefox_accounts_derived.fxa_users_services_first_seen_v2"
    ): FXA_SRC,
    user_id_target(
        table="firefox_accounts_derived.fxa_users_services_last_seen_v1"
    ): FXA_SRC,
    user_id_target(
        table="firefox_accounts_derived.fxa_users_services_last_seen_v2"
    ): FXA_SRC,
    user_id_target(
        table="firefox_accounts_derived.fxa_users_services_devices_daily_v1"
    ): FXA_SRC,
    user_id_target(
        table="firefox_accounts_derived.fxa_users_services_devices_first_seen_v1"
    ): FXA_SRC,
    user_id_target(
        table="firefox_accounts_derived.fxa_users_services_devices_last_seen_v1"
    ): FXA_SRC,
    context_id_target(
        table="contextual_services_stable.topsites_click_v1"
    ): CONTEXTUAL_SERVICES_SRC,
    context_id_target(
        table="contextual_services_stable.topsites_impression_v1"
    ): CONTEXTUAL_SERVICES_SRC,
    context_id_target(
        table="contextual_services_stable.quicksuggest_click_v1"
    ): CONTEXTUAL_SERVICES_SRC,
    context_id_target(
        table="contextual_services_stable.quicksuggest_impression_v1"
    ): CONTEXTUAL_SERVICES_SRC,
    # client association ping
    DeleteTarget(
        table="firefox_desktop_stable.fx_accounts_v1",
        field=("metrics.string.client_association_uid", GLEAN_CLIENT_ID),
    ): (FXA_UNHASHED_SRC, DESKTOP_GLEAN_SRC),
    # FxA on Glean
    DeleteTarget(
        table="accounts_backend_stable.events_v1",
        field="metrics.string.account_user_id_sha256",
    ): FXA_SRC,
    DeleteTarget(
        table="accounts_backend_stable.accounts_events_v1",
        field="metrics.string.account_user_id_sha256",
    ): FXA_SRC,
    DeleteTarget(
        table="accounts_backend_derived.events_stream_v1",
        field="metrics.string.account_user_id_sha256",
    ): FXA_SRC,
    DeleteTarget(
        table="accounts_backend_derived.users_services_daily_v1",
        field="user_id_sha256",
    ): FXA_SRC,
    DeleteTarget(
        table="accounts_backend_derived.users_services_last_seen_v1",
        field="user_id_sha256",
    ): FXA_SRC,
    DeleteTarget(
        table="accounts_frontend_stable.events_v1",
        field=("metrics.string.account_user_id_sha256", GLEAN_CLIENT_ID),
    ): (FXA_SRC, FXA_FRONTEND_GLEAN_SRC),
    DeleteTarget(
        table="accounts_frontend_stable.accounts_events_v1",
        field=("metrics.string.account_user_id_sha256", GLEAN_CLIENT_ID),
    ): (FXA_SRC, FXA_FRONTEND_GLEAN_SRC),
    DeleteTarget(
        table="accounts_frontend_derived.events_stream_v1",
        field=("metrics.string.account_user_id_sha256", CLIENT_ID),
    ): (FXA_SRC, FXA_FRONTEND_GLEAN_SRC),
    # legacy mobile
    DeleteTarget(
        table="telemetry_stable.core_v1",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v2",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v3",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v4",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v5",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v6",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v7",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v8",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v9",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v10",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.mobile_event_v1",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table=REGRETS_SRC.table,
        field="event_metadata.extension_installation_uuid",
    ): REGRETS_SRC,
    DeleteTarget(
        table="firefox_desktop_stable.user_characteristics_v1",
        field=USER_CHARACTERISTICS_ID,
    ): USER_CHARACTERISTICS_SRC,
    # tables in Glean derived datasets that use different sources than the find_glean_targets defaults
    client_id_target(table="firefox_desktop_derived.adclick_history_v1"): DESKTOP_SRC,
    client_id_target(table="firefox_desktop_derived.client_ltv_v1"): DESKTOP_SRC,
    client_id_target(table="firefox_desktop_derived.ltv_states_v1"): DESKTOP_SRC,
}

SEARCH_IGNORE_TABLES = {source.table for source in SOURCES}
SEARCH_IGNORE_TABLES |= {target.table for target in DELETE_TARGETS}
# these tables have a known user identifier, but do not yet have associated
# deletion requests, or do not keep data for older than 30 days
SEARCH_IGNORE_TABLES |= {
    target.table
    for target in [
        # glean migration
        glean_target(table="org_mozilla_fenix_nightly_stable.migration_v1"),
        glean_target(table="org_mozilla_fenix_stable.migration_v1"),
        glean_target(table="org_mozilla_fennec_aurora_stable.migration_v1"),
        glean_target(table="org_mozilla_firefox_beta_stable.migration_v1"),
        glean_target(table="org_mozilla_firefox_stable.migration_v1"),
        # pocket
        DeleteTarget(table="pocket_stable.fire_tv_events_v1", field=POCKET_ID),
        # mobile
        client_id_target(table="mobile_stable.activation_v1"),
        client_id_target(table="telemetry_stable.core_v1"),
        client_id_target(table="telemetry_stable.core_v2"),
        client_id_target(table="telemetry_stable.core_v3"),
        client_id_target(table="telemetry_stable.core_v4"),
        client_id_target(table="telemetry_stable.core_v5"),
        client_id_target(table="telemetry_stable.core_v6"),
        client_id_target(table="telemetry_stable.core_v7"),
        client_id_target(table="telemetry_stable.core_v8"),
        client_id_target(table="telemetry_stable.core_v9"),
        client_id_target(table="telemetry_stable.core_v10"),
        client_id_target(table="telemetry_stable.mobile_event_v1"),
        client_id_target(table="telemetry_stable.mobile_metrics_v1"),
        # internal
        client_id_target(table="eng_workflow_stable.build_v1"),
        # other
        DeleteTarget(table="telemetry_stable.pioneer_study_v4", field=PIONEER_ID),
    ]
}

# these fields should be ignored by search because they are not user identifiers
SEARCH_IGNORE_FIELDS = {
    # id is the source for document_id in these tables
    ("firefox_launcher_process_stable.launcher_process_failure_v1", ID),
    ("telemetry_stable.anonymous_v4", ID),
    ("telemetry_stable.optout_v4", ID),
    ("telemetry_stable.pre_account_v4", ID),
    ("telemetry_stable.prio_v4", ID),
}

# list of dataset_id.table_id to ignore in find_glean_targets function
GLEAN_IGNORE_LIST = {
    # deletion request table
    "firefox_desktop_derived.migration_esr_incorrect_deletion_request_v1",
    # subset of firefox_desktop_stable.pageload_v1 which doesn't have client ids
    "firefox_desktop_derived.pageload_1pct_v1",
    "firefox_desktop_derived.pageload_nightly_v1",
}


def find_glean_targets(
    pool: ThreadPool, client: bigquery.Client, project: str = SHARED_PROD
) -> DeleteIndex:
    """Return a dict like DELETE_TARGETS for glean tables.

    Note that dict values *must* be either DeleteSource or tuple[DeleteSource, ...],
    and other iterable types, e.g. list[DeleteSource] are not allowed or supported.
    """
    datasets = {dataset.dataset_id for dataset in client.list_datasets(project)}

    glean_stable_tables = [
        table
        for tables in pool.map(
            client.list_tables,
            [
                bigquery.DatasetReference(project, dataset_id)
                for dataset_id in datasets
                if dataset_id.endswith("_stable")
            ],
            chunksize=1,
        )
        for table in tables
        if table.labels.get("schema_id") == GLEAN_SCHEMA_ID
    ]

    channel_to_app_name = get_glean_channel_to_app_name_mapping()

    # create mapping of dataset -> (tables containing associated deletion requests)
    # construct values as tuples because that is what they must be in the return type
    sources: dict[str, tuple[DeleteSource, ...]] = defaultdict(tuple)
    app_names = set()
    source_doctype = "deletion_request"
    for table in glean_stable_tables:
        if table.table_id.startswith(source_doctype):
            source = DeleteSource(qualified_table_id(table), GLEAN_CLIENT_ID, project)

            channel_name = re.sub("_stable$", "", table.dataset_id)
            derived_dataset = channel_name + "_derived"
            app_name = channel_to_app_name.get(channel_name)

            # append to tuple to use every version of deletion request tables
            sources[table.dataset_id] += (source,)
            sources[derived_dataset] += (source,)

            # find the name of all apps that have a dataset of combined channels
            if app_name is not None and app_name != channel_name:
                app_names.add(app_name)
                sources[app_name + "_derived"] += (source,)

    # Use deletion request view containing all channels if found, otherwise use per-channel
    # tables as delete sources.  Some apps don't have views generated by glean_usage
    # because they are skipped in bqetl_project.yaml
    for app_name in app_names:
        try:
            source_view = client.get_table(f"{project}.{app_name}.{source_doctype}")
        except NotFound:
            pass
        else:
            source = DeleteSource(
                qualified_table_id(source_view), GLEAN_CLIENT_ID, project
            )
            sources[app_name + "_derived"] = (source,)

    glean_derived_tables = [
        table
        for table in pool.map(
            client.get_table,
            chain(
                *pool.starmap(
                    _list_tables,
                    [
                        (bigquery.DatasetReference(project, dataset_id), client)
                        for dataset_id in sources
                        if dataset_id.endswith("_derived")
                    ],
                    chunksize=1,
                )
            ),
            chunksize=1,
        )
        if table.table_type == "TABLE"
    ]

    # handle additional source for deletion requests for things like
    # https://bugzilla.mozilla.org/show_bug.cgi?id=1810236
    # table must contain client_id at the top level and be partitioned on
    # submission_timestamp
    derived_source_prefix = "additional_deletion_requests"
    for table in glean_derived_tables:
        if table.table_id.startswith(derived_source_prefix):
            source = DeleteSource(qualified_table_id(table), CLIENT_ID, project)
            channel_name = re.sub("_derived$", "", table.dataset_id)
            app_name = channel_to_app_name.get(channel_name)

            sources[channel_name + "_stable"] += (source,)
            sources[table.dataset_id] += (source,)

            if app_name is not None and app_name != channel_name:
                sources[app_name + "_derived"] += (source,)

    # skip tables already added to DELETE_TARGETS
    manually_added_tables = {target.table for target in DELETE_TARGETS.keys()}
    skipped_tables = manually_added_tables | GLEAN_IGNORE_LIST

    return {
        **{
            # glean stable tables that have a source
            DeleteTarget(
                table=qualified_table_id(table),
                # field must be repeated for each deletion source
                field=(GLEAN_CLIENT_ID,) * len(sources[table.dataset_id]),
            ): sources[table.dataset_id]
            for table in glean_stable_tables
            if table.dataset_id in sources
            and not table.table_id.startswith(source_doctype)
            # migration tables not yet supported
            and not table.table_id.startswith("migration")
            # skip tables with explicitly excluded client ids
            and table.labels.get("include_client_id", "true").lower() != "false"
            and qualified_table_id(table) not in skipped_tables
        },
        **{
            # glean derived tables that contain client_id
            DeleteTarget(
                table=qualified_table_id(table),
                # field must be repeated for each deletion source
                field=(CLIENT_ID,) * len(sources[table.dataset_id]),
            ): sources[table.dataset_id]
            for table in glean_derived_tables
            if any(field.name == CLIENT_ID for field in table.schema)
            and not table.table_id.startswith(derived_source_prefix)
            and qualified_table_id(table) not in skipped_tables
        },
        **{
            # glean derived tables that contain client_info.client_id but not client_id
            DeleteTarget(
                table=qualified_table_id(table),
                # field must be repeated for each deletion source
                field=(GLEAN_CLIENT_ID,) * len(sources[table.dataset_id]),
            ): sources[table.dataset_id]
            for table in glean_derived_tables
            if any(
                field.name == "client_info"
                and any(
                    [nested_field.name == "client_id" for nested_field in field.fields]
                )
                for field in table.schema
            )
            and all(field.name != CLIENT_ID for field in table.schema)
            and not table.table_id.startswith(derived_source_prefix)
            and qualified_table_id(table) not in skipped_tables
        },
    }


def get_glean_channel_to_app_name_mapping() -> Dict[str, str]:
    """Return a dict where key is the channel app id and the value is the shared app name.

    e.g. {
        "org_mozilla_firefox": "fenix",
        "org_mozilla_firefox_beta": "fenix",
        "org_mozilla_ios_firefox": "firefox_ios",
        "org_mozilla_ios_firefoxbeta": "firefox_ios",
    }
    """
    response = requests.get(GLEAN_APP_LISTINGS_URL)
    response.raise_for_status()

    app_listings = response.json()

    return {
        app["bq_dataset_family"]: app["app_name"]
        for app in app_listings
        if "bq_dataset_family" in app and "app_name" in app
    }


def _list_tables(
    dataset_ref: bigquery.DatasetReference,
    client: bigquery.Client,
) -> List:
    """Wrap bigquery list_tables and return an empty list for non-existent datasets.

    Intended to be used with thread pool map function. Some glean apps do not have
    derived datasets so this wrapper handles exceptions when listing them.
    """
    try:
        return list(client.list_tables(dataset_ref))
    except NotFound:
        return []


EXPERIMENT_ANALYSIS = "moz-fx-data-experiments"


def find_experiment_analysis_targets(
    client: bigquery.Client, project: str = EXPERIMENT_ANALYSIS
) -> DeleteIndex:
    """Return a dict like DELETE_TARGETS for experiment analysis tables.

    Note that dict values *must* be either DeleteSource or tuple[DeleteSource, ...],
    and other iterable types, e.g. list[DeleteSource] are not allowed or supported.
    """
    table_query = f"""
    SELECT
      table_schema AS dataset_id,
      table_name,
    FROM
      `{project}.region-us.INFORMATION_SCHEMA.TABLES`
    WHERE
      table_type = 'BASE TABLE'
      AND NOT STARTS_WITH(table_name, "statistics")
      AND ddl LIKE '%client_id STRING,%'
    """

    target_tables = client.query_and_wait(query=table_query)

    return {
        client_id_target(
            table=f"{table.dataset_id}.{table.table_name}", project=project
        ): DESKTOP_SRC
        for table in target_tables
    }


PIONEER_PROD = "moz-fx-data-pioneer-prod"


def find_pioneer_targets(
    pool: ThreadPool,
    client: bigquery.Client,
    project: str = PIONEER_PROD,
    study_projects: list[str] = [],
) -> DeleteIndex:
    """Return a dict like DELETE_TARGETS for Pioneer tables.

    Note that dict values *must* be either DeleteSource or tuple[DeleteSource, ...],
    and other iterable types, e.g. list[DeleteSource] are not allowed or supported.
    """

    def _has_nested_rally_id(field):
        """Check if any of the fields contains nested `metrics.uuid_rally_id`."""
        if field.name == "metrics" and field.field_type == "RECORD":
            uuid_field = next(filter(lambda f: f.name == "uuid", field.fields), None)
            if uuid_field and uuid_field.field_type == "RECORD":
                return any(field.name == "rally_id" for field in uuid_field.fields)
        return False

    def _get_tables_with_pioneer_id(dataset):
        tables_with_pioneer_id = []
        for table in client.list_tables(dataset):
            table_ref = client.get_table(table)
            if (
                any(field.name == PIONEER_ID for field in table_ref.schema)
                or any(field.name == RALLY_ID_TOP_LEVEL for field in table_ref.schema)
                or any(_has_nested_rally_id(field) for field in table_ref.schema)
            ) and table_ref.table_type != "VIEW":
                tables_with_pioneer_id.append(table_ref)
        return tables_with_pioneer_id

    def _get_client_id_field(table, deletion_request_view=False, study_name=None):
        """Determine which column should be used as client id for a given table."""
        if table.dataset_id.startswith("rally_"):
            # `rally_zero_one` is a special case where top-level rally_id is used
            # both in the ping tables and the deletion_requests view
            if table.dataset_id in ["rally_zero_one_stable", "rally_zero_one_derived"]:
                return RALLY_ID_TOP_LEVEL
            # deletion request views expose rally_id as a top-level field
            if deletion_request_view:
                return RALLY_ID_TOP_LEVEL
            else:
                return RALLY_ID
        elif table.dataset_id == "analysis":
            # Rally analysis tables do not have schemas specified upfront,
            # analysts might decide to use either nested or top-level rally_id.
            # Shared datasets, like attention stream, may also have derived
            # datasets with rally IDs
            # See https://github.com/mozilla-services/cloudops-infra/blob/master/projects/data-pioneer/tf/prod/envs/prod/study-projects/main.tf#L60-L67 # noqa
            if any(_has_nested_rally_id(field) for field in table.schema):
                return RALLY_ID
            elif any(field.name == RALLY_ID_TOP_LEVEL for field in table.schema):
                return RALLY_ID_TOP_LEVEL
            # Pioneer derived tables will have a PIONEER_ID
            elif any(field.name == PIONEER_ID for field in table.schema):
                return PIONEER_ID
            else:
                logging.error(f"Failed to find client_id field for {table}")
        else:
            return PIONEER_ID

    datasets = {
        dataset.reference
        for dataset in client.list_datasets(project)
        if dataset.reference.dataset_id.startswith("pioneer_")
        or dataset.reference.dataset_id.startswith("rally_")
    }
    # There should be a single stable and derived dataset per study
    stable_datasets = {dr for dr in datasets if dr.dataset_id.endswith("_stable")}
    derived_datasets = {dr for dr in datasets if dr.dataset_id.endswith("_derived")}

    stable_tables = [
        table
        for tables in pool.map(client.list_tables, stable_datasets, chunksize=1)
        for table in tables
    ]

    # Each derived deletion request view is a union of:
    # * corresponding deletion_request table from the _stable dataset
    # * uninstall_deletion pings from pioneer_core_stable dataset
    derived_deletion_request_views = [
        table
        for tables in pool.map(client.list_tables, derived_datasets, chunksize=1)
        for table in tables
        if (table.table_type == "VIEW" and table.table_id == "deletion_requests")
    ]

    # There is a derived dataset for each stable one
    # For simplicity when accessing this map later on, keys are changed to `_stable` here
    sources = {
        table.dataset_id.replace("_derived", "_stable"): DeleteSource(
            qualified_table_id(table),
            _get_client_id_field(table, deletion_request_view=True),
            project,
        )
        # dict comprehension will only keep the last value for a given key, so
        # sort by table_id to use the latest version
        for table in sorted(derived_deletion_request_views, key=lambda t: t.table_id)
    }

    # Dictionary mapping analysis dataset names to corresponding study names.
    # We expect analysis tables to be created only under `analysis` datasets
    # in study projects. These datasets are labeled with study names which
    # we use for discovering corresponding delete request tables later on.
    analysis_datasets = {}
    for project in study_projects:
        analysis_dataset = bigquery.DatasetReference(project, "analysis")
        labels = client.get_dataset(analysis_dataset).labels
        # study names in labels are not normalized (contain '-', not '_')
        study_name = labels.get("study_name")
        if study_name is None:
            logging.error(
                f"Dataset {analysis_dataset} does not have `study_name` label, skipping..."
            )
        else:
            analysis_datasets[analysis_dataset] = study_name

    return {
        **{
            # stable tables
            DeleteTarget(
                table=qualified_table_id(table),
                field=_get_client_id_field(table),
                project=PIONEER_PROD,
            ): sources[table.dataset_id]
            for table in stable_tables
            if not table.table_id.startswith("deletion_request_")
            and not table.table_id.startswith("pioneer_enrollment_")
            and not table.table_id.startswith("enrollment_")
            and not table.table_id.startswith("study_enrollment_")
            and not table.table_id.startswith("study_unenrollment_")
            and not table.table_id.startswith("unenrollment_")
        },
        **{
            # derived tables with pioneer_id
            DeleteTarget(
                table=qualified_table_id(table),
                field=_get_client_id_field(table),
                project=PIONEER_PROD,
            ): sources[table.dataset_id]
            for dataset in derived_datasets
            for table in _get_tables_with_pioneer_id(dataset)
        },
        **{
            # tables with pioneer_id located in study analysis projects
            DeleteTarget(
                table=qualified_table_id(table),
                field=_get_client_id_field(table, study_name=study),
                project=table.project,
            ): sources[study.replace("-", "_") + "_stable"]
            for dataset, study in analysis_datasets.items()
            for table in _get_tables_with_pioneer_id(dataset)
        },
    }



"""Common exception types."""


class BigQueryInsertError(Exception):
    """Wrap errors returned by BigQuery insertAll requests in an Exception."""

    @staticmethod
    def raise_if_present(errors):
        """Raise this exception if errors is truthy."""
        if errors:
            raise BigQueryInsertError(errors)



"""Functions that retrieve allow or block lists of probes."""


def get_etl_excluded_probes_quickfix(product):
    """Provide a static list of probes that must be excluded from aggregation.

    TODO: Change the function name and doc when final implementation is done.
    When so, this function will prob be turned into something like 'get_allowed_probes'
    and retrieve the list from an api call.
    """
    # See https://github.com/mozilla/glam/issues/1865
    forbidden_probes_by_product = {
        "fenix": {},
        "desktop": {"sqlite_store_open", "sqlite_store_query"},
    }
    return forbidden_probes_by_product[product]



from pathlib import Path

import pytest

from bigquery_etl.backfill.parse import (
    BACKFILL_FILE,
    DEFAULT_REASON,
    DEFAULT_WATCHER,
    Backfill,
    BackfillStatus,
)
from bigquery_etl.backfill.validate import (
    validate_default_reason,
    validate_default_watchers,
    validate_duplicate_entry_dates,
    validate_entries,
    validate_entries_are_sorted,
    validate_file,
)
from tests.backfill.test_parse_backfill import TEST_BACKFILL_1, TEST_BACKFILL_2

TEST_DIR = Path(__file__).parent.parent

VALID_REASON = "test_reason"
VALID_WATCHER = "test@example.org"


class TestValidateBackfill(object):

    def test_entries_duplicate_entry_dates_should_fail(self):
        backfills = [TEST_BACKFILL_1, TEST_BACKFILL_1]
        with pytest.raises(ValueError) as e:
            validate_duplicate_entry_dates(TEST_BACKFILL_1, backfills)

        assert "Duplicate backfill with entry date" in str(e.value)

    def test_valid_reason_should_pass(self):
        valid_backfill = Backfill(
            TEST_BACKFILL_1.entry_date,
            TEST_BACKFILL_1.start_date,
            TEST_BACKFILL_1.end_date,
            TEST_BACKFILL_1.excluded_dates,
            VALID_REASON,
            TEST_BACKFILL_1.watchers,
            TEST_BACKFILL_1.status,
        )

        validate_default_reason(valid_backfill)

    def test_validate_default_reason_should_fail(self):
        invalid_backfill = Backfill(
            TEST_BACKFILL_1.entry_date,
            TEST_BACKFILL_1.start_date,
            TEST_BACKFILL_1.end_date,
            TEST_BACKFILL_1.excluded_dates,
            DEFAULT_REASON,
            TEST_BACKFILL_1.watchers,
            TEST_BACKFILL_1.status,
        )

        with pytest.raises(ValueError) as e:
            validate_default_reason(invalid_backfill)

        assert "Default reason" in str(e.value)

    def test_validate_default_watcher_should_fail(self):
        TEST_BACKFILL_1.watchers = [DEFAULT_WATCHER]

        with pytest.raises(ValueError) as e:
            validate_default_watchers(TEST_BACKFILL_1)

        assert "Default watcher" in str(e.value)

    def test_entries_sorted(self):
        backfills = [TEST_BACKFILL_2, TEST_BACKFILL_1]
        validate_entries_are_sorted(backfills)

    def test_entries_not_sorted(self):
        backfills = [TEST_BACKFILL_1, TEST_BACKFILL_2]
        with pytest.raises(ValueError) as e:
            validate_entries_are_sorted(backfills)

        assert "Backfill entries are not sorted" in str(e.value)

    def test_validate_entries_duplicate_entry_with_initiate_status_should_fail(self):
        TEST_BACKFILL_1.watchers = [VALID_WATCHER]
        TEST_BACKFILL_1.reason = VALID_REASON
        TEST_BACKFILL_2.watchers = [VALID_WATCHER]
        TEST_BACKFILL_2.reason = VALID_REASON
        backfills = [TEST_BACKFILL_2, TEST_BACKFILL_1]
        with pytest.raises(ValueError) as e:
            validate_entries(backfills)
        assert (
            "Backfill entries cannot contain more than one entry with Initiate status"
            in str(e.value)
        )

    def test_validate_entries(self):
        TEST_BACKFILL_1.watchers = [VALID_WATCHER]
        TEST_BACKFILL_1.reason = VALID_REASON
        TEST_BACKFILL_2.watchers = [VALID_WATCHER]
        TEST_BACKFILL_2.reason = VALID_REASON
        TEST_BACKFILL_2.status = BackfillStatus.COMPLETE.value
        backfills = [TEST_BACKFILL_2, TEST_BACKFILL_1]
        validate_entries(backfills)

    def test_validate_file(self):
        backfill_file = TEST_DIR / "backfill" / "test_dir_valid" / BACKFILL_FILE
        validate_file(backfill_file)



import datetime

import pytest

from bigquery_etl.query_scheduling.formatters import (
    format_attr,
    format_date,
    format_optional_string,
    format_schedule_interval,
    format_timedelta,
)


class TestFormatters:
    def test_format_schedule_interval(self):
        assert "'@daily'" == format_schedule_interval("daily")
        assert "'@once'" == format_schedule_interval("once")
        assert "'* 1 * * *'" == format_schedule_interval("* 1 * * *")
        assert "'foo bar'" == format_schedule_interval("foo bar")

    def test_format_attr(self):
        assert format_attr({"a": 12}, "b", "format_schedule_interval") == {"a": 12}
        assert format_attr({"a": 12}, "a", "non_existing_formatter") == {"a": 12}
        assert format_attr(
            {"interval": "daily", "a": 12}, "interval", "format_schedule_interval"
        ) == {"interval": "'@daily'", "a": 12}

    def test_format_date(self):
        assert format_date(None) is None

        with pytest.raises(ValueError):
            assert format_date("March 12th 2020")

        with pytest.raises(ValueError):
            assert format_date("2020-13-01")

        assert format_date("2020-01-01") == datetime.datetime(2020, 1, 1, 0, 0)

    def test_format_timedelta(self):
        assert format_timedelta("string") == "string"

        with pytest.raises(TypeError):
            assert format_timedelta(123)

        assert format_timedelta("1h") == datetime.timedelta(seconds=3600)
        assert format_timedelta("-1h") == datetime.timedelta(seconds=-3600)
        assert format_timedelta("15m") == datetime.timedelta(seconds=900)
        assert format_timedelta("-15m") == datetime.timedelta(seconds=-900)
        assert format_timedelta("1h15m") == datetime.timedelta(seconds=4500)
        assert format_timedelta("-1h15m") == datetime.timedelta(seconds=-4500)
        assert format_timedelta("1s") == datetime.timedelta(seconds=1)
        assert format_timedelta("-1s") == datetime.timedelta(seconds=-1)
        assert format_timedelta("1h1m1s") == datetime.timedelta(seconds=3661)
        assert format_timedelta("-1h1m1s") == datetime.timedelta(seconds=-3661)
        assert format_timedelta("1d1h1m1s") == "1d1h1m1s"

    def test_format_optional_string(self):
        assert format_optional_string(None) == "None"

        with pytest.raises(TypeError):
            assert format_optional_string(123)

        assert format_optional_string("test") == "'test'"



"""Testing data for query."""

from itertools import product
from pathlib import Path
from uuid import uuid4

import yaml

ROOT = Path(__file__).parent

SUBMISSION_DATE = "2020-10-01"
APP_BUILD_ID = "2020100100"
OS = "Android"
PING_TYPE = "metrics"
MINIMUM_CLIENT_COUNT = 900
# Testing precondition: ping_type, os, and app_build_id must not be "*". See
# models.py under the scalar_bucket_counts parameters to see that sets fields
# are used in the static combinations. If these are set to "*", then they will
# be double counted...
CLIENTS_SCALAR_AGGREGATES = []
for i in range(MINIMUM_CLIENT_COUNT):
    data = {
        "client_id": str(uuid4()),
        "ping_type": PING_TYPE,
        "os": OS,
        "app_version": 84,
        "app_build_id": APP_BUILD_ID,
        "channel": "*",
        "scalar_aggregates": [
            {
                "metric": "places_manager_write_query_count",
                "metric_type": "counter",
                "key": "",
                "agg_type": "count",
                "value": 4.0,
            },
        ],
    }
    CLIENTS_SCALAR_AGGREGATES.append(data)


# we must generate the set of combinations. Each one of these have the same
# values though.

BASE_ROW = {
    "agg_type": "histogram",
    "app_build_id": "*",
    "app_version": 84,
    "bucket": "4.00",
    "bucket_count": 100,
    "channel": "*",
    "client_agg_type": "count",
    "count": 1,
    "key": "",
    "metric": "places_manager_write_query_count",
    "metric_type": "counter",
    "os": "*",
    "ping_type": "*",
    "range_max": 3.0,
    "range_min": 2.0,
}

# Didn't intend to code golf. This enumerates all of the "static combinations"
# by taking the cross product of all values. Each of these can take on a value
# from each of the clients above. Since each attribute combination has a single
# client, we do not have to change the "count" in the base row.
EXPECT = [
    {**BASE_ROW, **dict(zip(["bucket", "ping_type", "os", "app_build_id"], values))}
    for values in product(
        ["4.00", "8.00"], *zip([PING_TYPE, OS, APP_BUILD_ID], ["*"] * 3)
    )
]

prefix = "glam_etl"
tables = [
    (
        f"{prefix}.org_mozilla_fenix_glam_nightly__clients_scalar_aggregates_v1.yaml",
        CLIENTS_SCALAR_AGGREGATES,
    ),
    ("expect.yaml", EXPECT),
]
for name, data in tables:
    with (ROOT / name).open("w") as fp:
        yaml.dump(data, fp)



"""Format SQL."""

import glob
import os
import os.path
import sys
from functools import partial
from multiprocessing.pool import Pool
from pathlib import Path

from bigquery_etl.config import ConfigLoader
from bigquery_etl.format_sql.formatter import reformat  # noqa E402
from bigquery_etl.util.common import qualify_table_references_in_file


def skip_format():
    """Return a list of configured queries for which formatting should be skipped."""
    return [
        file
        for skip in ConfigLoader.get("format", "skip", fallback=[])
        for file in glob.glob(skip, recursive=True)
    ]


def skip_qualifying_references():
    """Return a list of configured queries where fully qualifying references should be skipped."""
    return [
        file
        for skip in ConfigLoader.get(
            "format", "skip_qualifying_references", fallback=[]
        )
        for file in glob.glob(skip, recursive=True)
    ]


def _format_path(check, path):
    query = Path(path).read_text()

    try:
        if not any([path.endswith(s) for s in skip_qualifying_references()]):
            fully_referenced_query = qualify_table_references_in_file(Path(path))
        else:
            fully_referenced_query = query
    except NotImplementedError:
        fully_referenced_query = query  # not implemented for scripts

    formatted = reformat(fully_referenced_query, trailing_newline=True)
    if query != formatted:
        if check:
            print(f"Needs reformatting: bqetl format {path}")
        else:
            with open(path, "w") as fp:
                fp.write(formatted)
            print(f"Reformatted: {path}")
        return 1
    else:
        return 0


def format(paths, check=False, parallelism=8):
    """Format SQL files."""
    if not paths:
        query = sys.stdin.read()
        formatted = reformat(query, trailing_newline=True)
        if not check:
            print(formatted, end="")
        if check and query != formatted:
            sys.exit(1)
    else:
        sql_files = []

        for path in paths:
            if os.path.isdir(path):
                sql_files.extend(
                    filepath
                    for dirpath, _, filenames in os.walk(path, followlinks=True)
                    for filename in filenames
                    if filename.endswith(".sql")
                    # skip tests/**/input.sql
                    and not (path.startswith("tests") and filename == "input.sql")
                    for filepath in [os.path.join(dirpath, filename)]
                    if not any([filepath.endswith(s) for s in skip_format()])
                )
            elif path:
                sql_files.append(path)
        if not sql_files:
            print("Error: no files were found to format")
            sys.exit(255)

        with Pool(parallelism) as pool:
            result = pool.map(partial(_format_path, check), sql_files)

        reformatted = sum(result)
        unchanged = len(sql_files) - reformatted
        print(
            ", ".join(
                f"{number} file{'s' if number > 1 else ''}"
                f"{' would be' if check else ''} {msg}"
                for number, msg in [
                    (reformatted, "reformatted"),
                    (unchanged, "left unchanged"),
                ]
                if number > 0
            )
            + "."
        )
        if check and reformatted:
            sys.exit(1)



from pathlib import Path

import pytest
from click.exceptions import BadParameter

from bigquery_etl.cli.utils import (
    is_authenticated,
    is_valid_dir,
    is_valid_file,
    is_valid_project,
    table_matches_patterns,
)

TEST_DIR = Path(__file__).parent.parent


class TestUtils:

    def test_is_valid_dir(self):
        with pytest.raises(BadParameter):
            assert is_valid_dir(None, None, "invalid")
        with pytest.raises(BadParameter):
            assert is_valid_dir(None, None, str(TEST_DIR / "data" / "dags.yaml"))
        assert is_valid_dir(None, None, str(TEST_DIR)) == str(TEST_DIR)

    def test_is_valid_file(self):
        with pytest.raises(BadParameter):
            assert is_valid_file(None, None, "invalid")
        with pytest.raises(BadParameter):
            assert is_valid_file(None, None, str(TEST_DIR))
        assert is_valid_file(None, None, str(TEST_DIR / "data" / "dags.yaml")) == str(
            TEST_DIR / "data" / "dags.yaml"
        )

    @pytest.mark.integration
    def test_is_authenticated(self):
        assert is_authenticated()

    def test_is_valid_project(self):
        assert is_valid_project(None, None, "mozfun")
        assert is_valid_project(None, None, "moz-fx-data-shared-prod")
        assert is_valid_project(None, None, "moz-fx-data-backfill-1")
        with pytest.raises(BadParameter):
            assert is_valid_project(None, None, "not-existing")

    def test_table_matches_patterns(self):
        assert not table_matches_patterns(
            table="telemetry_live.main_v4",
            pattern=["telemetry_live.main_v4", "telemetry_live.event_v4"],
            invert=True,
        )
        assert not table_matches_patterns(
            table="telemetry_live.main_v4",
            pattern="telemetry_live.main_v4",
            invert=True,
        )

        assert table_matches_patterns(
            table="telemetry_live.main_v4",
            pattern=["telemetry_live.first_shutdown_v4", "telemetry_live.event_v4"],
            invert=True,
        )
        assert table_matches_patterns(
            table="telemetry_live.main_v4",
            pattern="telemetry_live.event_v4",
            invert=True,
        )



"""Country code lookup generation."""

import os
from collections import Counter
from pathlib import Path

import click
import yaml
from jinja2 import Environment, FileSystemLoader

from bigquery_etl.cli.utils import use_cloud_function_option

FILE_PATH = Path(os.path.dirname(__file__))
TEMPLATES_PATH = FILE_PATH / "templates"


@click.command("generate")
@click.option(
    "--target_project",
    "--target-project",
    help="Which project the views should be written to.",
    default="moz-fx-data-shared-prod",
)
@click.option(
    "--output-dir",
    "--output_dir",
    help="The location to write to. Defaults to sql/.",
    default=Path("sql"),
    type=click.Path(file_okay=False),
)
@use_cloud_function_option
def generate(target_project, output_dir, use_cloud_function):
    """Generate a CSV that maps country aliases to a country code."""
    target_path = Path(f"{output_dir}/{target_project}/static/country_names_v1/")
    target_path.mkdir(parents=True, exist_ok=True)

    env = Environment(loader=FileSystemLoader(TEMPLATES_PATH))
    csv_template = env.get_template("data.csv")

    alias_path = FILE_PATH / "aliases.yaml"
    aliases_yaml = yaml.safe_load(alias_path.read_text())

    aliases_to_code = sorted(
        [
            (f'"{alias}"', country_code)
            for country_code, aliases in aliases_yaml.items()
            for alias in aliases
        ],
        key=lambda x: x[1],
    )

    # It's important that each alias maps to only one code,
    # we don't want to introduce fan-out in downstream queries.
    counter = Counter(alias for alias, _ in aliases_to_code)
    for alias, count in counter.items():
        if count > 1:
            raise ValueError(
                f"Multiple mappings found for {alias}, please update aliases.yaml."
            )

    with open(target_path / "data.csv", "w") as data_file:
        data_file.write(csv_template.render(aliases_to_code=aliases_to_code))



# Load libraries
import json
import os
from datetime import datetime, timedelta

import pandas as pd
import requests
from argparse import ArgumentParser
from google.cloud import bigquery
from google.cloud import storage

# Configs
brwsr_usg_configs = {
    "timeout_limit": 220,
    "device_types": ["DESKTOP", "MOBILE", "OTHER", "ALL"],
    "max_limit": 20,
    "operating_systems": [
        "ALL",
        "WINDOWS",
        "MACOSX",
        "IOS",
        "ANDROID",
        "CHROMEOS",
        "LINUX",
        "SMART_TV",
    ],
    "locations": [
        "ALL",
        "AU",
        "BE",
        "BG",
        "BR",
        "CA",
        "CH",
        "CZ",
        "DE",
        "DK",
        "EE",
        "ES",
        "FI",
        "FR",
        "GB",
        "HR",
        "IE",
        "IN",
        "JP",
        "KE",
        "IT",
        "CY",
        "LV",
        "LT",
        "LU",
        "HU",
        "MT",
        "MX",
        "NL",
        "AT",
        "PL",
        "PT",
        "RO",
        "SI",
        "SK",
        "US",
        "SE",
        "GR",
    ],
    "user_types": ["ALL", "LIKELY_AUTOMATED", "LIKELY_HUMAN"],
    "bucket": "gs://moz-fx-data-prod-external-data/",
    "bucket_no_gs": "moz-fx-data-prod-external-data",
    "results_stg_gcs_fpth": "cloudflare/browser_usage/RESULTS_STAGING/%s_results_from_%s_run.csv",
    "results_archive_gcs_fpath": "cloudflare/browser_usage/RESULTS_ARCHIVE/%s_results_from_%s_run.csv",
    "errors_stg_gcs_fpth": "cloudflare/browser_usage/ERRORS_STAGING/%s_errors_from_%s_run.csv",
    "errors_archive_gcs_fpath": "cloudflare/browser_usage/ERRORS_ARCHIVE/%s_errors_from_%s_run.csv",
    "gcp_project_id": "moz-fx-data-shared-prod",
    "gcp_conn_id": "google_cloud_shared_prod",
    "results_bq_stg_table": "moz-fx-data-shared-prod.cloudflare_derived.browser_results_stg",
    "errors_bq_stg_table": "moz-fx-data-shared-prod.cloudflare_derived.browser_errors_stg",
}

#Load the Cloudflare API Token
cloudflare_api_token = os.getenv("CLOUDFLARE_AUTH_TOKEN")

# Define a function to move a GCS object then delete the original
def move_blob(bucket_name, blob_name, destination_bucket_name, destination_blob_name):
    """Moves a blob from one bucket to another with a new name."""
    storage_client = storage.Client()
    source_bucket = storage_client.bucket(bucket_name)
    source_blob = source_bucket.blob(blob_name)
    destination_bucket = storage_client.bucket(destination_bucket_name)
    destination_generation_match_precondition = None

    blob_copy = source_bucket.copy_blob(
        source_blob,
        destination_bucket,
        destination_blob_name,
        if_generation_match=destination_generation_match_precondition,
    )
    source_bucket.delete_blob(blob_name)

    print(
        "Blob {} in bucket {} moved to blob {} in bucket {}.".format(
            source_blob.name,
            source_bucket.name,
            blob_copy.name,
            destination_bucket.name,
        )
    )


# Define function to generate API call based on configs passed
def generate_browser_api_call(
    strt_dt, end_dt, device_type, location, op_system, user_typ, limit
):
    """Create the API url based on the input parameters."""
    user_type_string = "" if user_typ == "ALL" else f"&botClass={user_typ}"
    location_string = "" if location == "ALL" else f"&location={location}"
    op_system_string = "" if op_system == "ALL" else f"&os={op_system}"
    device_type_string = "" if device_type == "ALL" else f"&deviceType={device_type}"
    browser_api_url = f"https://api.cloudflare.com/client/v4/radar/http/top/browsers?dateStart={strt_dt}T00:00:00.000Z&dateEnd={end_dt}T00:00:00.000Z{device_type_string}{location_string}{op_system_string}{user_type_string}&limit={limit}&format=json"
    return browser_api_url


# Define function to pull browser data from the cloudflare API
def get_browser_data(date_of_interest, auth_token):
    """Pull browser data for each combination of the configs from the Cloudflare API, always runs with a lag of 4 days."""
    # Calculate start date and end date
    logical_dag_dt = date_of_interest
    logical_dag_dt_as_date = datetime.strptime(logical_dag_dt, "%Y-%m-%d").date()
    start_date = logical_dag_dt_as_date - timedelta(days=4)
    end_date = start_date + timedelta(days=1)
    print("Start Date: ", start_date)
    print("End Date: ", end_date)

    # Configure request headers
    bearer_string = f"Bearer {auth_token}"
    headers = {"Authorization": bearer_string}
    limit = brwsr_usg_configs["max_limit"]

    # Initialize the empty results and errors dataframes
    browser_results_df = pd.DataFrame(
        {
            "StartTime": [],
            "EndTime": [],
            "DeviceType": [],
            "Location": [],
            "UserType": [],
            "Browser": [],
            "OperatingSystem": [],
            "PercentShare": [],
            "ConfLevel": [],
            "Normalization": [],
            "LastUpdated": [],
        }
    )

    browser_errors_df = pd.DataFrame(
        {
            "StartTime": [],
            "EndTime": [],
            "Location": [],
            "UserType": [],
            "DeviceType": [],
            "OperatingSystem": [],
        }
    )

    # Loop through the combinations
    for device_type in brwsr_usg_configs["device_types"]:
        for loc in brwsr_usg_configs["locations"]:
            for os in brwsr_usg_configs["operating_systems"]:
                for user_type in brwsr_usg_configs["user_types"]:
                    curr_combo = f"DeviceType: {device_type}, Loc: {loc}, OS: {os}, UserType: {user_type}, Limit: {limit}"
                    print(curr_combo)

                    # Generate the URL & call the API
                    brwsr_usg_api_url = generate_browser_api_call(
                        start_date, end_date, device_type, loc, os, user_type, limit
                    )
                    try:
                        response = requests.get(
                            brwsr_usg_api_url,
                            headers=headers,
                            timeout=brwsr_usg_configs["timeout_limit"],
                        )
                        response_json = json.loads(response.text)

                        # if the response was successful, get the result and append it to the results dataframe
                        if response_json["success"] is True:
                            # Save the results to GCS
                            result = response_json["result"]
                            confidence_level = result["meta"]["confidenceInfo"]["level"]
                            normalization = result["meta"]["normalization"]
                            last_updated = result["meta"]["lastUpdated"]
                            startTime = result["meta"]["dateRange"][0]["startTime"]
                            endTime = result["meta"]["dateRange"][0]["endTime"]
                            data = result["top_0"]
                            browser_lst = []
                            browser_share_lst = []

                            for browser in data:
                                browser_lst.append(browser["name"])
                                browser_share_lst.append(browser["value"])

                            new_browser_results_df = pd.DataFrame(
                                {
                                    "StartTime": [startTime] * len(browser_lst),
                                    "EndTime": [endTime] * len(browser_lst),
                                    "DeviceType": [device_type] * len(browser_lst),
                                    "Location": [loc] * len(browser_lst),
                                    "UserType": [user_type] * len(browser_lst),
                                    "Browser": browser_lst,
                                    "OperatingSystem": [os] * len(browser_lst),
                                    "PercentShare": browser_share_lst,
                                    "ConfLevel": [confidence_level] * len(browser_lst),
                                    "Normalization": [normalization] * len(browser_lst),
                                    "LastUpdated": [last_updated] * len(browser_lst),
                                }
                            )
                            browser_results_df = pd.concat(
                                [browser_results_df, new_browser_results_df]
                            )

                        # If there were errors, save them to the errors dataframe
                        else:
                            new_browser_error_df = pd.DataFrame(
                                {
                                    "StartTime": [start_date],
                                    "EndTime": [end_date],
                                    "Location": [loc],
                                    "UserType": [user_type],
                                    "DeviceType": [device_type],
                                    "OperatingSystem": [os],
                                }
                            )
                            browser_errors_df = pd.concat(
                                [browser_errors_df, new_browser_error_df]
                            )
                    except:
                        new_browser_error_df = pd.DataFrame(
                            {
                                "StartTime": [start_date],
                                "EndTime": [end_date],
                                "Location": [loc],
                                "UserType": [user_type],
                                "DeviceType": [device_type],
                                "OperatingSystem": [os],
                            }
                        )
                        browser_errors_df = pd.concat(
                            [browser_errors_df, new_browser_error_df]
                        )

    # LOAD RESULTS & ERRORS TO STAGING GCS
    result_fpath = brwsr_usg_configs["bucket"] + brwsr_usg_configs[
        "results_stg_gcs_fpth"
    ] % (start_date, logical_dag_dt)
    error_fpath = brwsr_usg_configs["bucket"] + brwsr_usg_configs[
        "errors_stg_gcs_fpth"
    ] % (start_date, logical_dag_dt)
    browser_results_df.to_csv(result_fpath, index=False)
    browser_errors_df.to_csv(error_fpath, index=False)
    print("Wrote errors to: ", error_fpath)
    print("Wrote results to: ", result_fpath)

    # Return a summary to the console
    len_results = str(len(browser_results_df))
    len_errors = str(len(browser_errors_df))
    result_summary = f"# Result Rows: {len_results}; # of Error Rows: {len_errors}"
    return result_summary


def main():
    """Call the API, save data to GCS, load to BQ staging, delete & load to BQ gold"""
    parser = ArgumentParser(description=__doc__)
    parser.add_argument("--date", required=True)
    parser.add_argument("--cloudflare_api_token", default=cloudflare_api_token)
    parser.add_argument("--project", default=brwsr_usg_configs["gcp_project_id"])
    parser.add_argument("--dataset", default="cloudflare_derived")

    args = parser.parse_args()
    print("Running for date: ")
    print(args.date)

    # STEP 1 - Pull the data from the API, save results & errors to GCS staging area
    result_summary = get_browser_data(args.date, args.cloudflare_api_token)
    print("result_summary")
    print(result_summary)

    # Create a bigquery client
    client = bigquery.Client(args.project)

    result_uri = brwsr_usg_configs["bucket"] + brwsr_usg_configs[
        "results_stg_gcs_fpth"
    ] % (datetime.strptime(args.date, "%Y-%m-%d").date() - timedelta(days=4), args.date)
    error_uri = brwsr_usg_configs["bucket"] + brwsr_usg_configs[
        "errors_stg_gcs_fpth"
    ] % (datetime.strptime(args.date, "%Y-%m-%d").date() - timedelta(days=4), args.date)
    print("result_uri")
    print(result_uri)

    print("error_uri")
    print(error_uri)

    # STEP 2 - Copy the result data from GCS staging to BQ staging table
    load_result_csv_to_bq_stg_job = client.load_table_from_uri(
        result_uri,
        brwsr_usg_configs["results_bq_stg_table"],
        job_config=bigquery.LoadJobConfig(
            create_disposition="CREATE_IF_NEEDED",
            write_disposition="WRITE_TRUNCATE",
            schema=[
                {"name": "StartTime", "type": "TIMESTAMP", "mode": "REQUIRED"},
                {"name": "EndTime", "type": "TIMESTAMP", "mode": "REQUIRED"},
                {"name": "DeviceType", "type": "STRING", "mode": "NULLABLE"},
                {"name": "Location", "type": "STRING", "mode": "NULLABLE"},
                {"name": "UserType", "type": "STRING", "mode": "NULLABLE"},
                {"name": "Browser", "type": "STRING", "mode": "NULLABLE"},
                {"name": "OperatingSystem", "type": "STRING", "mode": "NULLABLE"},
                {"name": "PercentShare", "type": "NUMERIC", "mode": "NULLABLE"},
                {"name": "ConfLevel", "type": "STRING", "mode": "NULLABLE"},
                {"name": "Normalization", "type": "STRING", "mode": "NULLABLE"},
                {"name": "LastUpdated", "type": "TIMESTAMP", "mode": "NULLABLE"},
            ],
            skip_leading_rows=1,
            source_format=bigquery.SourceFormat.CSV,
        ),
    )
    load_result_csv_to_bq_stg_job.result()
    result_bq_stg_tbl = client.get_table(brwsr_usg_configs["results_bq_stg_table"])
    print("Loaded {} rows to results staging.".format(result_bq_stg_tbl.num_rows))

    # STEP 3 - Copy the error data from GCS staging to BQ staging table
    load_error_csv_to_bq_stg_job = client.load_table_from_uri(
        error_uri,
        brwsr_usg_configs["errors_bq_stg_table"],
        job_config=bigquery.LoadJobConfig(
            create_disposition="CREATE_IF_NEEDED",
            write_disposition="WRITE_TRUNCATE",
            schema=[
                {"name": "StartTime", "type": "DATE", "mode": "REQUIRED"},
                {"name": "EndTime", "type": "DATE", "mode": "REQUIRED"},
                {"name": "Location", "type": "STRING", "mode": "NULLABLE"},
                {"name": "UserType", "type": "STRING", "mode": "NULLABLE"},
                {"name": "DeviceType", "type": "STRING", "mode": "NULLABLE"},
                {"name": "OperatingSystem", "type": "STRING", "mode": "NULLABLE"},
            ],
            skip_leading_rows=1,
            source_format=bigquery.SourceFormat.CSV,
        ),
    )

    load_error_csv_to_bq_stg_job.result()
    error_bq_stg_tbl = client.get_table(brwsr_usg_configs["errors_bq_stg_table"])
    print("Loaded {} rows to errors staging.".format(error_bq_stg_tbl.num_rows))

    # STEP 4 - Delete results from gold for this day, if there are any already (so if rerun, no issues will occur)
    del_exstng_gold_res_for_date = f"""DELETE FROM `moz-fx-data-shared-prod.cloudflare_derived.browser_usage_v1` WHERE dte = DATE_SUB('{args.date}', INTERVAL 4 DAY)  """
    del_gold_res_job = client.query(del_exstng_gold_res_for_date)
    del_gold_res_job.result()
    print("Deleted anything already existing for this date from results gold")

    # STEP 5 - Delete errors from gold for this day, if there are any already (so if rerun, no issues will occur)
    del_exstng_gold_err_for_date = f"""DELETE FROM `moz-fx-data-shared-prod.cloudflare_derived.browser_usage_errors_v1` WHERE dte = DATE_SUB('{args.date}', INTERVAL 4 DAY) """
    del_gold_err_job = client.query(del_exstng_gold_err_for_date)
    del_gold_err_job.result()
    print("Deleted anything already existing for this date from errors gold")

    # STEP 6 - Load results from stage to gold
    browser_usg_stg_to_gold_query = f""" INSERT INTO `moz-fx-data-shared-prod.cloudflare_derived.browser_usage_v1`
SELECT
CAST(StartTime as date) AS dte,
DeviceType AS device_type,
Location AS location,
UserType AS user_type,
Browser AS browser,
OperatingSystem AS operating_system,
PercentShare AS percent_share,
Normalization AS normalization,
LastUpdated AS last_updated_ts
FROM `moz-fx-data-shared-prod.cloudflare_derived.browser_results_stg`
WHERE CAST(StartTime as date) = DATE_SUB('{args.date}', INTERVAL 4 DAY) """
    load_res_to_gold = client.query(browser_usg_stg_to_gold_query)
    load_res_to_gold.result()

    # STEP 7 - Load errors from stage to gold
    browser_usg_errors_stg_to_gold_query = f""" INSERT INTO `moz-fx-data-shared-prod.cloudflare_derived.browser_usage_errors_v1`
SELECT
CAST(StartTime as date) AS dte,
Location AS location,
UserType AS user_type,
DeviceType AS device_type,
OperatingSystem AS operating_system
FROM `moz-fx-data-shared-prod.cloudflare_derived.browser_errors_stg`
WHERE CAST(StartTime as date) = DATE_SUB('{args.date}', INTERVAL 4 DAY) """
    load_err_to_gold = client.query(browser_usg_errors_stg_to_gold_query)
    load_err_to_gold.result()

    # STEP 8 - Copy the result CSV from stage to archive, then delete from stage

    # Calculate the fpaths we will use ahead of time
    result_stg_fpath = brwsr_usg_configs["results_stg_gcs_fpth"] % (
        datetime.strptime(args.date, "%Y-%m-%d").date() - timedelta(days=4),
        args.date,
    )
    result_archive_fpath = brwsr_usg_configs["results_archive_gcs_fpath"] % (
        datetime.strptime(args.date, "%Y-%m-%d").date() - timedelta(days=4),
        args.date,
    )
    move_blob(
        brwsr_usg_configs["bucket_no_gs"],
        result_stg_fpath,
        brwsr_usg_configs["bucket_no_gs"],
        result_archive_fpath,
    )

    # STEP 9 - Copy the error CSV from stage to archive, then delete from stage
    error_stg_fpath = brwsr_usg_configs["errors_stg_gcs_fpth"] % (
        datetime.strptime(args.date, "%Y-%m-%d").date() - timedelta(days=4),
        args.date,
    )
    error_archive_fpath = brwsr_usg_configs["errors_archive_gcs_fpath"] % (
        datetime.strptime(args.date, "%Y-%m-%d").date() - timedelta(days=4),
        args.date,
    )
    move_blob(
        brwsr_usg_configs["bucket_no_gs"],
        error_stg_fpath,
        brwsr_usg_configs["bucket_no_gs"],
        error_archive_fpath,
    )


if __name__ == "__main__":
    main()



"""Generating and run baseline_clients_daily queries for Glean apps."""

from sql_generators.glean_usage.common import GleanTable

BASELINE_DAILY_TABLE_ID = "baseline_clients_daily_v1"
PREFIX = "daily"


class BaselineClientsDailyTable(GleanTable):
    """Represents generated baseline_clients_daily table."""

    def __init__(self):
        """Initialize baseline_clients_daily table."""
        GleanTable.__init__(self)
        self.target_table_id = BASELINE_DAILY_TABLE_ID
        self.prefix = PREFIX
        self.custom_render_kwargs = {}



from bigquery_etl.util.mozfun_docs_functions import (
    clean_result,
    clean_sql_text,
    get_input,
    get_mozfun_parameters,
    get_output,
)


class TestUDFFunctions:
    class TestCleanSQLText:
        def test_clean_sql_text_multiple_whitespaces(self):
            assert (
                clean_sql_text("multiple  white      spaces    in   line")
                == "multiple white spaces in line"
            )

        def test_clean_sql_text_line_breaks(self):
            assert clean_sql_text("all\nthe\nline\nbreaks") == "all the line breaks"

        def test_clean_sql_text_comment_whole_line(self):
            assert clean_sql_text("""something\n-- comment\nelse""") == "something else"

        def test_clean_sql_text_comment_in_line(self):
            assert clean_sql_text("some text -- end line comment") == "some text"

    class TestCleanResult:
        def test_clean_result_no_brackets(self):
            assert clean_result("nothing to do") == "nothing to do"

        def test_clean_result_struct(self):
            assert clean_result("STRUCT < foo STRING >") == "STRUCT<foo STRING>"

    class TestGetInput:
        def test_get_input_single_input(self):
            assert (
                get_input(
                    "CREATE OR REPLACE FUNCTION test_dataset.test_udf(input1 INT64)"
                )
                == "input1 INT64"
            )

        def test_get_input_multiple_inputs(self):
            assert (
                get_input(
                    "CREATE OR REPLACE FUNCTION test_thing.test_udf(input1 INT64, input2 FLOAT64, length INT64)"
                )
                == "input1 INT64, input2 FLOAT64, length INT64"
            )

        def test_get_input_array_struct_input(self):
            assert (
                get_input(
                    """CREATE OR REPLACE FUNCTION glam.histogram_fill_buckets_dirichlet(input_map ARRAY<STRUCT<key STRING, value FLOAT64>>)"""
                )
                == "input_map ARRAY<STRUCT<key STRING, value FLOAT64>>"
            )

        def test_get_input_multiline_input(self):
            assert (
                get_input(
                    """CREATE OR REPLACE FUNCTION glam.histogram_fill_buckets_dirichlet(
                input_map ARRAY<STRUCT<key STRING, value FLOAT64>>,
                buckets ARRAY<STRING>,
                total_users INT64
                )"""
                )
                == "input_map ARRAY<STRUCT<key STRING, value FLOAT64>>, buckets ARRAY<STRING>, total_users INT64"
            )

        def test_get_input_multiline_input_with_udf_body(self):
            assert (
                get_input(
                    """CREATE OR REPLACE FUNCTION glam.histogram_fill_buckets_dirichlet(
                input_map ARRAY<STRUCT<key STRING, value FLOAT64>>,
                buckets ARRAY<STRING>,
                total_users INT64
                ) AS
                SELECT 1"""
                )
                == "input_map ARRAY<STRUCT<key STRING, value FLOAT64>>, buckets ARRAY<STRING>, total_users INT64"
            )

        def test_get_input_none_input(self):
            assert get_input("some other text") is None

    class TestGetOutput:
        def test_get_output_single_output(self):
            assert get_output("RETURNS BOOLEAN AS (") == "BOOLEAN"

        def test_get_output_array_struct_output(self):
            assert (
                get_output(
                    "total_users INT64)RETURNS ARRAY<STRUCT<key STRING, value FLOAT64>> AS ("
                )
                == "ARRAY<STRUCT<key STRING, value FLOAT64>>"
            )

        def test_get_output_with_udf_body_single_line(self):
            assert (
                get_output(
                    """CREATE OR REPLACE FUNCTION glam.histogram_fill_buckets_dirichlet(
                        input_map ARRAY<STRUCT<key STRING, value FLOAT64>>,
                        buckets ARRAY<STRING>,
                        total_users INT64
                        ) RETURNS BOOLEAN AS SELECT 1"""
                )
                == "BOOLEAN"
            )

        def test_get_output_with_udf_body_multiline(self):
            assert (
                get_output(
                    """CREATE OR REPLACE FUNCTION glam.histogram_fill_buckets_dirichlet(
                    input_map ARRAY<STRUCT<key STRING, value FLOAT64>>,
                    buckets ARRAY<STRING>,
                    total_users INT64
                    ) RETURNS BOOLEAN AS
                    SELECT 1"""
                )
                == "BOOLEAN"
            )

        def test_get_output_from_cast(self):
            assert (
                get_output(
                    """CREATE OR REPLACE FUNCTION bits28.from_string(s STRING) AS (
                IF(
                    REGEXP_CONTAINS(s, r"^[01]{1,28}$"),
                    (
                    SELECT
                        SUM(CAST(c AS INT64) << (LENGTH(s) - 1 - bit))
                    FROM
                        UNNEST(SPLIT(s, '')) AS c
                        WITH OFFSET bit
                    ),
                    ERROR(FORMAT("bits28_from_string expects a string of up to 28 0's and 1's but got: %s", s))
                )
                );"""
                )
                == "INT64"
            )

        def test_get_output_multiple_matches(self):
            assert (
                get_output(
                    """CREATE OR REPLACE FUNCTION foo.bar(s STRING) RETURNS INT64 AS (
                        SELECT CAST(s AS FLOAT64)
                    )
                    """
                )
                == "INT64"
            )

        def test_get_output_without_as(self):
            # note that the odd string formatting below is intentional and spacing affects test results
            assert (
                get_output(
                    """
                    CREATE OR REPLACE FUNCTION norm.get_windows_info(os_version STRING)
RETURNS STRUCT<name STRING, version_name STRING, version_number DECIMAL, build_number INT64>
LANGUAGE js AS r\"\"\"
  const test = "test";
  return test;
\"\"\";
                    """
                )
                == "STRUCT<name STRING, version_name STRING, version_number DECIMAL, build_number INT64>"
            )

        def test_get_output_to_string(self):
            assert (
                get_output(
                    """CREATE OR REPLACE FUNCTION bits28.to_string(bits INT64) AS (
                        (
                            SELECT
                            STRING_AGG(CAST(bits >> bit & 0x1 AS STRING), '' ORDER BY bit DESC)
                            FROM
                            UNNEST(GENERATE_ARRAY(0, 27)) AS bit
                        )
                        );
                    """
                )
                == "STRING"
            )

        def test_get_output_create_count_steps_query(self):
            assert (
                get_output(
                    """CREATE OR REPLACE PROCEDURE
                        event_analysis.create_count_steps_query(
                            project STRING,
                            dataset STRING,
                            events ARRAY<STRUCT<category STRING, event_name STRING>>,
                            OUT sql STRING
                        )
                    """
                )
                == "sql STRING"
            )

    class TestGetMozfunParameters:
        def test_get_mozfun_parameters_1(self):
            sql_text = """CREATE OR REPLACE FUNCTION test_dataset.test_udf(input1 INT64, input2 FLOAT64) RETURNS INT64 AS SELECT 4"""

            input_part, output_part = get_mozfun_parameters(sql_text)
            assert input_part == "input1 INT64, input2 FLOAT64"
            assert output_part == "INT64"

        def test_get_mozfun_parameters_2(self):
            sql_text = """/*

                Return a boolean indicating if any bits are set in the specified range of a bit pattern.

                The start_offset must be zero or a negative number indicating an offset from
                the rightmost bit in the pattern.

                n_bits is the number of bits to consider, counting right from the bit at start_offset.

                See detailed docs for the bits28 suite of functions:
                https://docs.telemetry.mozilla.org/cookbooks/clients_last_seen_bits.html#udf-reference

                */
                CREATE OR REPLACE FUNCTION bits28.active_in_range(bits INT64, start_offset INT64, n_bits INT64)
                RETURNS BOOLEAN AS (
                CASE
                    WHEN start_offset > 0
                    THEN ERROR(
                        FORMAT(
                            'start_offset must be <= 0 but was %i in call bits28_active_in_range(%i, %i, %i)',
                            start_offset,
                            bits,
                            start_offset,
                            n_bits
                        )
                        )
                    WHEN n_bits > (1 - start_offset)
                    THEN ERROR(
                        FORMAT(
                            'Reading %i bits from starting_offset %i exceeds end of bit pattern in call bits28_active_in_range(%i, %i, %i)',
                            n_bits,
                            start_offset,
                            bits,
                            start_offset,
                            n_bits
                        )
                        )
                    ELSE BIT_COUNT(bits28.range(bits, start_offset, n_bits)) > 0
                END
                );

                -- Tests
                SELECT
                assert.true(bits28.active_in_range(1 << 10, -13, 7)),
                assert.false(bits28.active_in_range(1 << 10, -6, 7)),
                assert.true(bits28.active_in_range(1, 0, 1)),
                assert.false(bits28.active_in_range(0, 0, 1));
            """

            input_part, output_part = get_mozfun_parameters(sql_text)
            assert input_part == "bits INT64, start_offset INT64, n_bits INT64"
            assert output_part == "BOOLEAN"

        def test_get_mozfun_parameters_3(self):
            sql_text = """-- udf_bucket
                CREATE OR REPLACE FUNCTION glam.histogram_bucket_from_value(buckets ARRAY<STRING>, val FLOAT64)
                RETURNS FLOAT64 AS (
                -- Bucket `value` into a histogram with min_bucket, max_bucket and num_buckets
                (
                    SELECT
                    MAX(CAST(bucket AS FLOAT64))
                    FROM
                    UNNEST(buckets) AS bucket
                    WHERE
                    val >= CAST(bucket AS FLOAT64)
                )
                );

                --Tests
                SELECT
                assert.equals(2.0, glam.histogram_bucket_from_value(["1", "2", "3"], 2.333)),
                assert.equals(NULL, glam.histogram_bucket_from_value(["1"], 0.99)),
                assert.equals(0.0, glam.histogram_bucket_from_value(["0", "1"], 0.99)),
                assert.equals(0.0, glam.histogram_bucket_from_value(["1", "0"], 0.99)),
            """

            input_part, output_part = get_mozfun_parameters(sql_text)
            assert input_part == "buckets ARRAY<STRING>, val FLOAT64"
            assert output_part == "FLOAT64"

        def test_get_mozfun_parameters_4(self):
            sql_text = """-- udf_fill_buckets
                CREATE OR REPLACE FUNCTION glam.histogram_fill_buckets_dirichlet(
                input_map ARRAY<STRUCT<key STRING, value FLOAT64>>,
                buckets ARRAY<STRING>,
                total_users INT64
                )
                RETURNS ARRAY<STRUCT<key STRING, value FLOAT64>> AS (
                -- Given a MAP `input_map`, fill in any missing keys with value `0.0` and
                -- transform values to estimate a Dirichlet distribution
                ARRAY(
                    SELECT AS STRUCT
                    key,
                    -- Dirichlet distribution density for each bucket in a histogram.
                    -- Given client level {k1: p1, k2:p2, ... , kK: pK}
                    -- where p’s are client proportions, and p1, p2, ... pK sum to 1,
                    -- k1, k2, ... , kK are the buckets, and K is the total number of buckets.
                    -- returns an array [] around a struct {k : v} such that
                    -- [{k1: (P1+1/K) / (N_reporting+1), k2:(P2+1/K) / (N_reporting+1), ...}]
                    -- where (capital) P1 is the sum of p1s, P2 for p2s, etc.:
                    --   P1 = (p1_client1 + p1_client2 + ... + p1_clientN) & 0 < P1 < N
                    -- and capital K is again the total number of buckets.
                    -- For more information, please see:
                    -- https://docs.google.com/document/d/1ipy1oFIKDvHr3R6Ku0goRjS11R1ZH1z2gygOGkSdqUg
                    SAFE_DIVIDE(
                        COALESCE(e.value, 0.0) + SAFE_DIVIDE(1, ARRAY_LENGTH(buckets)),
                        total_users + 1
                    ) AS value
                    FROM
                    UNNEST(buckets) AS key
                    LEFT JOIN
                    UNNEST(input_map) AS e
                    ON key = e.key
                    ORDER BY
                    SAFE_CAST(key AS FLOAT64)
                )
                );

                SELECT
                -- fill in 1 with a value of 0
                assert.array_equals(
                    ARRAY<STRUCT<key STRING, value FLOAT64>>[
                    ("0", (1 + (1 / 3)) / 3),
                    ("1", 1 / 9),
                    ("2", (2 + (1 / 3)) / 3)
                    ],
                    glam.histogram_fill_buckets_dirichlet(
                    ARRAY<STRUCT<key STRING, value FLOAT64>>[("0", 1.0), ("2", 2.0)],
                    ["0", "1", "2"],
                    2
                    )
                ),
                -- only keep values in specified in buckets
                assert.array_equals(
                    ARRAY<STRUCT<key STRING, value FLOAT64>>[("0", (1 + 1) / 3)],
                    glam.histogram_fill_buckets_dirichlet(
                    ARRAY<STRUCT<key STRING, value FLOAT64>>[("0", 1.0), ("2", 1.0)],
                    ["0"],
                    2
                    )
                ),
                -- keys may not non-integer values, so ordering is not well defined for strings
                assert.array_equals(
                    ARRAY<STRUCT<key STRING, value FLOAT64>>[
                    ("foo", (1 + (1 / 2)) / 3),
                    ("bar", (1 + (1 / 2)) / 3)
                    ],
                    glam.histogram_fill_buckets_dirichlet(
                    ARRAY<STRUCT<key STRING, value FLOAT64>>[("foo", 1.0), ("bar", 1.0)],
                    ["foo", "bar"],
                    2
                    )
                ),
                -- but ordering is guaranteed for integers/floats
                assert.array_equals(
                    ARRAY<STRUCT<key STRING, value FLOAT64>>[("2", (1 + (1 / 2)) / 3), ("11", (1 + (1 / 2)) / 3)],
                    glam.histogram_fill_buckets_dirichlet(
                    ARRAY<STRUCT<key STRING, value FLOAT64>>[("11", 1.0), ("2", 1.0)],
                    ["11", "2"],
                    2
                    )
                )
            """
            assert get_mozfun_parameters(sql_text) == (
                "input_map ARRAY<STRUCT<key STRING, value FLOAT64>>, buckets ARRAY<STRING>, total_users INT64",
                "ARRAY<STRUCT<key STRING, value FLOAT64>>",
            )

        def test_get_mozfun_parameters_5(self):
            sql_text = """-- udf_get_values
                CREATE OR REPLACE FUNCTION glam.map_from_array_offsets(
                required ARRAY<FLOAT64>,
                `values` ARRAY<FLOAT64>
                )
                RETURNS ARRAY<STRUCT<key STRING, value FLOAT64>> AS (
                (
                    SELECT
                    ARRAY_AGG(
                        STRUCT<key STRING, value FLOAT64>(
                        CAST(key AS STRING),
                        `values`[OFFSET(SAFE_CAST(key AS INT64))]
                        )
                        ORDER BY
                        key
                    )
                    FROM
                    UNNEST(required) AS key
                )
                );

                SELECT
                assert.array_equals(
                    ARRAY<STRUCT<key STRING, value FLOAT64>>[("0", 2.0), ("2", 4.0)],
                    glam.map_from_array_offsets([0.0, 2.0], [2.0, 3.0, 4.0])
                ),
                -- required ordered
                assert.array_equals(
                    ARRAY<STRUCT<key STRING, value FLOAT64>>[("0", 2.0), ("2", 4.0)],
                    glam.map_from_array_offsets([2.0, 0.0], [2.0, 3.0, 4.0])
                ),
                -- intended use-case, for approx_quantiles
                assert.array_equals(
                    ARRAY<STRUCT<key STRING, value FLOAT64>>[("25", 1.0), ("50", 2.0), ("75", 3.0)],
                    glam.map_from_array_offsets(
                    [25.0, 50.0, 75.0],
                    (
                        SELECT
                        ARRAY_AGG(CAST(y * 1.0 AS FLOAT64) ORDER BY i)
                        FROM
                        UNNEST((SELECT APPROX_QUANTILES(x, 100) FROM UNNEST([1, 2, 3]) AS x)) AS y
                        WITH OFFSET i
                    )
                    )
                )
            """

            assert get_mozfun_parameters(sql_text) == (
                "required ARRAY<FLOAT64>, `values` ARRAY<FLOAT64>",
                "ARRAY<STRUCT<key STRING, value FLOAT64>>",
            )

        def test_get_mozfun_parameters_6(self):
            sql_text = """/*

                Convert a string representing individual bits into an INT64.

                Implementation based on https://stackoverflow.com/a/51600210/1260237

                See detailed docs for the bits28 suite of functions:
                https://docs.telemetry.mozilla.org/cookbooks/clients_last_seen_bits.html#udf-reference

                */
                CREATE OR REPLACE FUNCTION bits28.from_string(s STRING) AS (
                IF(
                    REGEXP_CONTAINS(s, r"^[01]{1,28}$"),
                    (
                    SELECT
                        SUM(CAST(c AS INT64) << (LENGTH(s) - 1 - bit))
                    FROM
                        UNNEST(SPLIT(s, '')) AS c
                        WITH OFFSET bit
                    ),
                    ERROR(FORMAT("bits28_from_string expects a string of up to 28 0's and 1's but got: %s", s))
                )
                );

                -- Tests
                SELECT
                assert.equals(1, bits28.from_string('1')),
                assert.equals(1, bits28.from_string('01')),
                assert.equals(1, bits28.from_string('0000000000000000000000000001')),
                assert.equals(2, bits28.from_string('10')),
                assert.equals(5, bits28.from_string('101'));
            """
            assert get_mozfun_parameters(sql_text) == ("s STRING", "INT64")

        def test_get_mozfun_parameters_procedure_1(self):
            sql_text = """ CREATE OR REPLACE PROCEDURE
                event_analysis.create_count_steps_query(
                    project STRING,
                    dataset STRING,
                    events ARRAY<STRUCT<category STRING, event_name STRING>>,
                    OUT sql STRING
                )
                BEGIN
                DECLARE i INT64 DEFAULT 1;

                DECLARE event STRUCT<category STRING, event_name STRING>;

                DECLARE event_filter STRING;

                DECLARE event_filters ARRAY<STRING> DEFAULT[];

                WHILE
                    i <= ARRAY_LENGTH(events)
                DO
                    SET event = events[ORDINAL(i)];

                    SET event_filter = CONCAT(
                    '(category = "',
                    event.category,
                    '"',
                    ' AND event = "',
                    event.event_name,
                    '")'
                    );

                    SET event_filters = ARRAY_CONCAT(event_filters, [event_filter]);

                    SET i = i + 1;
                END WHILE;

                SET sql = CONCAT(
                    '\n  SELECT',
                    '\n    event_analysis.aggregate_match_strings(ARRAY_AGG(event_analysis.event_index_to_match_string(index))) AS count_regex',
                    '\n  FROM',
                    '\n    `',
                    project,
                    '`.',
                    dataset,
                    '.event_types',
                    '\n  WHERE',
                    '\n    ',
                    ARRAY_TO_STRING(event_filters, ' OR ')
                );
                END;

                -- Tests
                BEGIN
                DECLARE result_sql STRING;

                DECLARE expect STRING DEFAULT \"\"\"
                SELECT
                    event_analysis.aggregate_match_strings(ARRAY_AGG(event_analysis.event_index_to_match_string(index))) AS count_regex
                FROM
                    `moz-fx-data-shared-prod`.org_mozilla_firefox.event_types
                WHERE
                    (category = "collections" AND event = "saved") OR (category = "collections" AND event = "tabs_added")
                \"\"\";

                CALL event_analysis.create_count_steps_query(
                    'moz-fx-data-shared-prod',
                    'org_mozilla_firefox',
                    [
                    STRUCT('collections' AS category, 'saved' AS event_name),
                    STRUCT('collections' AS category, 'tabs_added' AS event_name)
                    ],
                    result_sql
                );

                SELECT
                    assert.sql_equals(expect, result_sql);
                END;

            """
            assert get_mozfun_parameters(sql_text) == (
                "project STRING, dataset STRING, events ARRAY<STRUCT<category STRING, event_name STRING>>",
                "sql STRING",
            )

        def test_get_mozfun_parameters_from_procedure_1(self):
            assert get_mozfun_parameters(
                "CREATE OR REPLACE PROCEDURE foo.bar(s STRING, b BOOLEAN) AS ..."
            ) == ("s STRING, b BOOLEAN", None)

        def test_get_mozfun_parameters_from_procedure_2(self):
            assert get_mozfun_parameters(
                "CREATE OR REPLACE PROCEDURE foo.bar(s STRING, b BOOLEAN, OUT sql STRING) AS ..."
            ) == ("s STRING, b BOOLEAN", "sql STRING")

        def test_get_mozfun_parameters_from_days_since_seen(self):
            test_text = """/*

                Return the position of the rightmost set bit in an INT64 bit pattern.

                To determine this position, we take a bitwise AND of the bit pattern and
                its complement, then we determine the position of the bit via base-2 logarithm;
                see https://stackoverflow.com/a/42747608/1260237

                See detailed docs for the bits28 suite of functions:
                https://docs.telemetry.mozilla.org/cookbooks/clients_last_seen_bits.html#udf-reference

                */
                CREATE OR REPLACE FUNCTION bits28.days_since_seen(bits INT64) AS (
                CAST(SAFE.LOG(bits & -bits, 2) AS INT64)
                );

                SELECT
                assert.null(bits28.days_since_seen(0)),
                assert.equals(0, bits28.days_since_seen(1)),
                assert.equals(3, bits28.days_since_seen(8)),
                assert.equals(0, bits28.days_since_seen(8 + 1))
            """

            assert get_mozfun_parameters(sql_text=test_text) == ("bits INT64", "INT64")

        def test_get_mozfun_parameters_from_to_string(self):
            test_text = """/*

                Convert an INT64 field into a 28-character string representing the individual bits.

                Implementation based on https://stackoverflow.com/a/51600210/1260237

                See detailed docs for the bits28 suite of functions:
                https://docs.telemetry.mozilla.org/cookbooks/clients_last_seen_bits.html#udf-reference

                */
                CREATE OR REPLACE FUNCTION bits28.to_string(bits INT64) AS (
                (
                    SELECT
                    STRING_AGG(CAST(bits >> bit & 0x1 AS STRING), '' ORDER BY bit DESC)
                    FROM
                    UNNEST(GENERATE_ARRAY(0, 27)) AS bit
                )
                );

                -- Tests
                SELECT
                assert.equals('0100000000000000000000000010', bits28.to_string((1 << 1) | (1 << 26)))
            """
            assert get_mozfun_parameters(sql_text=test_text) == ("bits INT64", "STRING")

        def test_get_mozfun_parameters_from_retention(self):
            test_text = """CREATE OR REPLACE FUNCTION bits28.retention(bits INT64, submission_date DATE) AS (
                STRUCT(
                    STRUCT(
                    submission_date AS metric_date,
                    bits28.active_in_range(bits, 0, 1) AS active_on_metric_date
                    ) AS day_0,
                    STRUCT(
                    DATE_SUB(submission_date, INTERVAL 6 DAY) AS metric_date,
                    bits28.active_in_range(bits, -6, 1) AS active_on_metric_date,
                    bits28.active_in_range(bits, -6, 7) AS active_in_week_0,
                    bits28.active_in_range(bits, -5, 6) AS active_in_week_0_after_metric_date
                    ) AS day_6,
                    STRUCT(
                    DATE_SUB(submission_date, INTERVAL 13 DAY) AS metric_date,
                    bits28.active_in_range(bits, -13, 1) AS active_on_metric_date,
                    bits28.active_in_range(bits, -13, 7) AS active_in_week_0,
                    bits28.active_in_range(bits, -12, 6) AS active_in_week_0_after_metric_date,
                    bits28.active_in_range(bits, -6, 7) AS active_in_week_1
                    ) AS day_13,
                    STRUCT(
                    DATE_SUB(submission_date, INTERVAL 20 DAY) AS metric_date,
                    bits28.active_in_range(bits, -20, 1) AS active_on_metric_date,
                    bits28.active_in_range(bits, -20, 7) AS active_in_week_0,
                    bits28.active_in_range(bits, -19, 6) AS active_in_week_0_after_metric_date,
                    bits28.active_in_range(bits, -13, 7) AS active_in_week_1,
                    bits28.active_in_range(bits, -6, 7) AS active_in_week_2
                    ) AS day_20,
                    STRUCT(
                    DATE_SUB(submission_date, INTERVAL 27 DAY) AS metric_date,
                    bits28.active_in_range(bits, -27, 1) AS active_on_metric_date,
                    bits28.active_in_range(bits, -27, 7) AS active_in_week_0,
                    bits28.active_in_range(bits, -26, 6) AS active_in_week_0_after_metric_date,
                    bits28.active_in_range(bits, -20, 7) AS active_in_week_1,
                    bits28.active_in_range(bits, -13, 7) AS active_in_week_2,
                    bits28.active_in_range(bits, -6, 7) AS active_in_week_3
                    ) AS day_27
                )
                );
            """
            assert get_mozfun_parameters(sql_text=test_text) == (
                "bits INT64, submission_date DATE",
                None,
            )

        def test_get_mozfun_parameters_from_create_count_steps_query(self):
            test_text = """CREATE OR REPLACE PROCEDURE
                event_analysis.create_count_steps_query(
                    project STRING,
                    dataset STRING,
                    events ARRAY<STRUCT<category STRING, event_name STRING>>,
                    OUT sql STRING
                )
                BEGIN
            """
            assert get_mozfun_parameters(sql_text=test_text) == (
                "project STRING, dataset STRING, events ARRAY<STRUCT<category STRING, event_name STRING>>",
                "sql STRING",
            )

        def test_get_mozfun_parameters_from_scrub_apple_receipt(self):
            test_text = """CREATE OR REPLACE FUNCTION iap.scrub_apple_receipt(apple_receipt ANY TYPE)
                RETURNS STRUCT<
                environment STRING,
                active_period STRUCT<
                    -- *_date fields deprecated in favor of TIMESTAMP fields
                    start_date DATE,
                    end_date DATE,
                    start_time TIMESTAMP,
                    end_time TIMESTAMP,
                    `interval` STRING,
                    interval_count INT64
                >,
                trial_period STRUCT<start_time TIMESTAMP, end_time TIMESTAMP>
                > AS (
            """
            assert get_mozfun_parameters(sql_text=test_text) == (
                "apple_receipt ANY TYPE",
                "STRUCT<environment STRING, active_period STRUCT<start_date DATE, end_date DATE, start_time TIMESTAMP, \
end_time TIMESTAMP, `interval` STRING, interval_count INT64>, trial_period STRUCT<start_time TIMESTAMP, end_time TIMESTAMP>>",
            )



"""Generate active users aggregates per app."""

import os
from enum import Enum
from pathlib import Path

import click
from jinja2 import Environment, FileSystemLoader

from bigquery_etl.cli.utils import use_cloud_function_option
from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.util.common import render, write_sql

THIS_PATH = Path(os.path.dirname(__file__))
TABLE_NAME_DELETION_REQUEST = "active_users_aggregates_deletion_request"


class Browsers(Enum):
    """Enumeration with browser names and equivalent dataset names."""

    # The initial implementation is a prototype for some mobile browsers.
    # Desktop is excluded due to the bigger size of data.
    # Focus Android is excluded due to its higher complexity.
    firefox_desktop = "Firefox Desktop"
    fenix = "Fenix"
    focus_ios = "Focus iOS"
    firefox_ios = "Firefox iOS"
    klar_ios = "Klar iOS"


@click.command()
@click.option(
    "--output-dir",
    "--output_dir",
    help="Output directory generated SQL is written to",
    type=click.Path(file_okay=False),
    default="sql",
)
@click.option(
    "--target-project",
    "--target_project",
    help="Google Cloud project ID",
    default="moz-fx-data-shared-prod",
)
@use_cloud_function_option
def generate(target_project, output_dir, use_cloud_function):
    """Generate per-app queries, views and metadata for active users aggregates.

    As calculated for the client_ids who have sent a deletion request ping.
    """
    env = Environment(loader=FileSystemLoader(str(THIS_PATH / "templates")))
    mobile_deletion_request_query_template = env.get_template(
        "mobile_deletion_request_query.sql"
    )
    desktop_deletion_request_query_template = env.get_template(
        "desktop_deletion_request_query.sql"
    )
    metadata_deletion_request_template = "metadata_deletion_request.yaml"
    output_dir = Path(output_dir) / target_project

    for browser in Browsers:
        if browser.name == "firefox_desktop":
            write_sql(
                output_dir=output_dir,
                full_table_id=f"{target_project}.{browser.name}_derived.{TABLE_NAME_DELETION_REQUEST}_v1",
                basename="query.sql",
                sql=reformat(
                    desktop_deletion_request_query_template.render(
                        project_id=target_project,
                        app_value=browser.value,
                        app_name=browser.name,
                    )
                ),
                skip_existing=False,
            )
        else:
            write_sql(
                output_dir=output_dir,
                full_table_id=f"{target_project}.{browser.name}_derived.{TABLE_NAME_DELETION_REQUEST}_v1",
                basename="query.sql",
                sql=reformat(
                    mobile_deletion_request_query_template.render(
                        project_id=target_project,
                        app_value=browser.value,
                        app_name=browser.name,
                    )
                ),
                skip_existing=False,
            )

        write_sql(
            output_dir=output_dir,
            full_table_id=f"{target_project}.{browser.name}_derived.{TABLE_NAME_DELETION_REQUEST}_v1",
            basename="metadata.yaml",
            sql=render(
                metadata_deletion_request_template,
                template_folder=THIS_PATH / "templates",
                app_value=browser.value,
                app_name=browser.name,
                parameters='["end_date:DATE:{{macros.ds_add(ds, 27)}}", "start_date:DATE:{{macros.ds_add(ds, 27-28*4)}}"]',
                format=False,
            ),
            skip_existing=False,
        )



# Load libraries
import json
import os
from datetime import datetime, timedelta

import pandas as pd
import requests
from argparse import ArgumentParser
from google.cloud import bigquery
from google.cloud import storage

# Configurations
os_usg_configs = {
    "timeout_limit": 500,
    "device_types": ["DESKTOP", "MOBILE", "OTHER", "ALL"],
    "locations": [
        "ALL",
        "AU",
        "BE",
        "BG",
        "BR",
        "CA",
        "CH",
        "CZ",
        "DE",
        "DK",
        "EE",
        "ES",
        "FI",
        "FR",
        "GB",
        "HR",
        "IE",
        "IN",
        "IT",
        "JP",
        "KE",
        "CY",
        "LV",
        "LT",
        "LU",
        "HU",
        "MT",
        "MX",
        "NL",
        "AT",
        "PL",
        "PT",
        "RO",
        "SI",
        "SK",
        "US",
        "SE",
        "GR",
    ],
    "bucket": "gs://moz-fx-data-prod-external-data/",
    "bucket_no_gs": "moz-fx-data-prod-external-data",
    "results_stg_gcs_fpth": "cloudflare/os_usage/RESULTS_STAGING/%s_results_from_%s_run.csv",
    "results_archive_gcs_fpth": "cloudflare/os_usage/RESULTS_ARCHIVE/%s_results_from_%s_run.csv",
    "errors_stg_gcs_fpth": "cloudflare/os_usage/ERRORS_STAGING/%s_results_from_%s_run.csv",
    "errors_archive_gcs_fpth": "cloudflare/os_usage/ERRORS_ARCHIVE/%s_results_from_%s_run.csv",
    "gcp_project_id": "moz-fx-data-shared-prod",
    "gcp_conn_id": "google_cloud_shared_prod",
    "results_bq_stg_table": "moz-fx-data-shared-prod.cloudflare_derived.os_results_stg",
    "errors_bq_stg_table": "moz-fx-data-shared-prod.cloudflare_derived.os_errors_stg",
}

#Load the Cloudflare API Token
cloudflare_api_token = os.getenv("CLOUDFLARE_AUTH_TOKEN")

# Define a function to move a GCS object then delete the original
def move_blob(bucket_name, blob_name, destination_bucket_name, destination_blob_name):
    """Moves a blob from one bucket to another with a new name."""
    storage_client = storage.Client()
    source_bucket = storage_client.bucket(bucket_name)
    source_blob = source_bucket.blob(blob_name)
    destination_bucket = storage_client.bucket(destination_bucket_name)
    destination_generation_match_precondition = None

    blob_copy = source_bucket.copy_blob(
        source_blob,
        destination_bucket,
        destination_blob_name,
        if_generation_match=destination_generation_match_precondition,
    )
    source_bucket.delete_blob(blob_name)

    print(
        "Blob {} in bucket {} moved to blob {} in bucket {}.".format(
            source_blob.name,
            source_bucket.name,
            blob_copy.name,
            destination_bucket.name,
        )
    )


# Function to configure the API URL
def generate_os_timeseries_api_call(strt_dt, end_dt, agg_int, location, device_type):
    """Generate the API call for Operating System Usage Data."""
    if location == "ALL" and device_type == "ALL":
        os_usage_api_url = f"https://api.cloudflare.com/client/v4/radar/http/timeseries_groups/os?dateStart={strt_dt}T00:00:00.000Z&dateEnd={end_dt}T00:00:00.000Z&format=json&aggInterval={agg_int}"
    elif location != "ALL" and device_type == "ALL":
        os_usage_api_url = f"https://api.cloudflare.com/client/v4/radar/http/timeseries_groups/os?dateStart={strt_dt}T00:00:00.000Z&dateEnd={end_dt}T00:00:00.000Z&location={location}&format=json&aggInterval={agg_int}"
    elif location == "ALL" and device_type != "ALL":
        os_usage_api_url = f"https://api.cloudflare.com/client/v4/radar/http/timeseries_groups/os?dateStart={strt_dt}T00:00:00.000Z&dateEnd={end_dt}T00:00:00.000Z&deviceType={device_type}&format=json&aggInterval={agg_int}"
    else:
        os_usage_api_url = f"https://api.cloudflare.com/client/v4/radar/http/timeseries_groups/os?dateStart={strt_dt}T00:00:00.000Z&dateEnd={end_dt}T00:00:00.000Z&location={location}&deviceType={device_type}&format=json&aggInterval={agg_int}"
    return os_usage_api_url


def get_os_usage_data(date_of_interest, auth_token):
    """Pull OS usage data from the Cloudflare API and save errors & results to GCS."""
    # Calculate start date and end date
    logical_dag_dt = date_of_interest
    logical_dag_dt_as_date = datetime.strptime(logical_dag_dt, "%Y-%m-%d").date()
    start_date = logical_dag_dt_as_date - timedelta(days=4)
    end_date = start_date + timedelta(days=1)
    print("Start Date: ", start_date)
    print("End Date: ", end_date)

    # Configure request headers
    bearer_string = f"Bearer {auth_token}"
    headers = {"Authorization": bearer_string}

    # Initialize the empty results & errors dataframe
    result_df = pd.DataFrame(
        {
            "Timestamps": [],
            "OS": [],
            "Location": [],
            "DeviceType": [],
            "Share": [],
            "ConfidenceLevel": [],
            "AggrInterval": [],
            "Normalization": [],
            "LastUpdatedTS": [],
        }
    )

    # Initialize an errors dataframe
    errors_df = pd.DataFrame(
        {"StartTime": [], "EndTime": [], "Location": [], "DeviceType": []}
    )

    # Go through all combinations, submit API requests
    for device_type in os_usg_configs["device_types"]:
        for loc in os_usg_configs["locations"]:
            print("Device Type: ", device_type)
            print("Loc: ", loc)

            # Generate the URL with given parameters
            os_usage_api_url = generate_os_timeseries_api_call(
                start_date, end_date, "1d", loc, device_type
            )
            try:
                # Call the API and save the response as JSON
                response = requests.get(
                    os_usage_api_url,
                    headers=headers,
                    timeout=os_usg_configs["timeout_limit"],
                )
                response_json = json.loads(response.text)

                # If response was successful, get the result
                if response_json["success"] is True:
                    result = response_json["result"]
                    # Parse metadata
                    conf_lvl = result["meta"]["confidenceInfo"]["level"]
                    aggr_intvl = result["meta"]["aggInterval"]
                    nrmlztn = result["meta"]["normalization"]
                    lst_upd = result["meta"]["lastUpdated"]
                    data_dict = result["serie_0"]

                    for key, val in data_dict.items():
                        if key != 'timestamps':
                            new_result_df = pd.DataFrame(
                                {
                                    "Timestamps": data_dict["timestamps"],
                                    "OS": [key] * len(val),
                                    "Location": [loc] * len(val),
                                    "DeviceType": [device_type] * len(val),
                                    "Share": val,
                                    "ConfidenceLevel": [conf_lvl] * len(val),
                                    "AggrInterval": [aggr_intvl] * len(val),
                                    "Normalization": [nrmlztn] * len(val),
                                    "LastUpdatedTS": [lst_upd] * len(val),
                                }
                            )
                            result_df = pd.concat([result_df, new_result_df])

                # If response was not successful, get the errors
                else:
                    # errors = response_json["errors"]  # Maybe add to capture, right now not using this
                    new_errors_df = pd.DataFrame(
                        {
                            "StartTime": [start_date],
                            "EndTime": [end_date],
                            "Location": [loc],
                            "DeviceType": [device_type],
                        }
                    )
                    errors_df = pd.concat([errors_df, new_errors_df])
            except:
                new_errors_df = pd.DataFrame(
                    {
                        "StartTime": [start_date],
                        "EndTime": [end_date],
                        "Location": [loc],
                        "DeviceType": [device_type],
                    }
                )
                errors_df = pd.concat([errors_df, new_errors_df])

    result_fpath = os_usg_configs["bucket"] + os_usg_configs["results_stg_gcs_fpth"] % (
        start_date,
        logical_dag_dt,
    )
    errors_fpath = os_usg_configs["bucket"] + os_usg_configs["errors_stg_gcs_fpth"] % (
        start_date,
        logical_dag_dt,
    )

    result_df.to_csv(result_fpath, index=False)
    errors_df.to_csv(errors_fpath, index=False)
    print("Wrote errors to: ", errors_fpath)
    print("Wrote results to: ", result_fpath)

    # Write a summary to the logs
    len_results = str(len(result_df))
    len_errors = str(len(errors_df))
    result_summary = f"# Result Rows: {len_results}; # of Error Rows: {len_errors}"
    return result_summary


def main():
    """Call the API, save data to GCS, load to BQ staging, delete & load to BQ gold"""
    parser = ArgumentParser(description=__doc__)
    parser.add_argument("--date", required=True)
    parser.add_argument("--cloudflare_api_token", default=cloudflare_api_token)
    parser.add_argument("--project", default="moz-fx-data-shared-prod")
    parser.add_argument("--dataset", default="cloudflare_derived")

    args = parser.parse_args()
    print("Running for date: ")
    print(args.date)

    # STEP 1 - Pull the data from the API, save results & errors to GCS staging area
    result_summary = get_os_usage_data(args.date, args.cloudflare_api_token)
    print("result_summary")
    print(result_summary)

    # Create a bigquery client
    client = bigquery.Client(args.project)

    result_uri = os_usg_configs["bucket"] + os_usg_configs["results_stg_gcs_fpth"] % (
        datetime.strptime(args.date, "%Y-%m-%d").date() - timedelta(days=4),
        args.date,
    )
    error_uri = os_usg_configs["bucket"] + os_usg_configs["errors_stg_gcs_fpth"] % (
        datetime.strptime(args.date, "%Y-%m-%d").date() - timedelta(days=4),
        args.date,
    )
    print("result_uri")
    print(result_uri)

    print("error_uri")
    print(error_uri)

    # STEP 2 - Copy the result data from GCS staging to BQ staging table
    load_result_csv_to_bq_stg_job = client.load_table_from_uri(
        result_uri,
        os_usg_configs["results_bq_stg_table"],
        job_config=bigquery.LoadJobConfig(
            create_disposition="CREATE_IF_NEEDED",
            write_disposition="WRITE_TRUNCATE",
            schema=[
                {"name": "Timestamps", "type": "TIMESTAMP", "mode": "REQUIRED"},
                {"name": "OS", "type": "STRING", "mode": "NULLABLE"},
                {"name": "Location", "type": "STRING", "mode": "NULLABLE"},
                {"name": "DeviceType", "type": "STRING", "mode": "NULLABLE"},
                {"name": "Share", "type": "NUMERIC", "mode": "NULLABLE"},
                {"name": "ConfidenceLevel", "type": "STRING", "mode": "NULLABLE"},
                {"name": "AggrInterval", "type": "STRING", "mode": "NULLABLE"},
                {"name": "Normalization", "type": "STRING", "mode": "NULLABLE"},
                {"name": "LastUpdatedTS", "type": "TIMESTAMP", "mode": "NULLABLE"},
            ],
            skip_leading_rows=1,
            source_format=bigquery.SourceFormat.CSV,
        ),
    )

    load_result_csv_to_bq_stg_job.result()
    result_bq_stg_tbl = client.get_table(os_usg_configs["results_bq_stg_table"])
    print("Loaded {} rows to results staging.".format(result_bq_stg_tbl.num_rows))

    # STEP 3 - Copy the error data from GCS staging to BQ staging table
    load_error_csv_to_bq_stg_job = client.load_table_from_uri(
        error_uri,
        os_usg_configs["errors_bq_stg_table"],
        job_config=bigquery.LoadJobConfig(
            create_disposition="CREATE_IF_NEEDED",
            write_disposition="WRITE_TRUNCATE",
            schema=[
                {"name": "StartDate", "type": "DATE", "mode": "REQUIRED"},
                {"name": "EndDate", "type": "DATE", "mode": "REQUIRED"},
                {"name": "Location", "type": "STRING", "mode": "NULLABLE"},
                {"name": "DeviceType", "type": "STRING", "mode": "NULLABLE"},
            ],
            skip_leading_rows=1,
            source_format=bigquery.SourceFormat.CSV,
        ),
    )

    load_error_csv_to_bq_stg_job.result()
    error_bq_stg_tbl = client.get_table(os_usg_configs["errors_bq_stg_table"])
    print("Loaded {} rows to errors staging.".format(error_bq_stg_tbl.num_rows))

    # STEP 4 - Delete results from gold for this day, if there are any already (so if rerun, no issues will occur)
    del_exstng_gold_res_for_date = f"""DELETE FROM `moz-fx-data-shared-prod.cloudflare_derived.os_usage_v1` WHERE dte = DATE_SUB('{args.date}', INTERVAL 4 DAY)  """
    del_gold_res_job = client.query(del_exstng_gold_res_for_date)
    del_gold_res_job.result()
    print("Deleted anything already existing for this date from results gold")

    # STEP 5 - Delete errors from gold for this day, if there are any already (so if rerun, no issues will occur)
    del_exstng_gold_err_for_date = f"""DELETE FROM `moz-fx-data-shared-prod.cloudflare_derived.os_usage_errors_v1` WHERE dte = DATE_SUB('{args.date}', INTERVAL 4 DAY) """
    del_gold_err_job = client.query(del_exstng_gold_err_for_date)
    del_gold_err_job.result()
    print("Deleted anything already existing for this date from errors gold")

    # STEP 6 - Load results from stage to gold # NEED TO UPDATE THIS STILL
    os_usg_stg_to_gold_query = f""" INSERT INTO `moz-fx-data-shared-prod.cloudflare_derived.os_usage_v1`
SELECT 
CAST(Timestamps AS date) AS dte,
OS AS os,
Location AS location,
DeviceType AS device_type,
Share AS os_share,
Normalization AS normalization_type,
LastUpdatedTS AS last_updated_ts
FROM `moz-fx-data-shared-prod.cloudflare_derived.os_results_stg`
WHERE CAST(Timestamps as date) = DATE_SUB('{args.date}', INTERVAL 4 DAY) """
    load_res_to_gold = client.query(os_usg_stg_to_gold_query)
    load_res_to_gold.result()

    # STEP 7 - Load errors from stage to gold
    os_usg_errors_stg_to_gold_query = f""" INSERT INTO `moz-fx-data-shared-prod.cloudflare_derived.os_usage_errors_v1`
SELECT
StartDate AS dte,
Location AS location,
DeviceType AS device_type
FROM `moz-fx-data-shared-prod.cloudflare_derived.os_errors_stg`
WHERE StartDate = DATE_SUB('{args.date}', INTERVAL 4 DAY) """
    load_err_to_gold = client.query(os_usg_errors_stg_to_gold_query)
    load_err_to_gold.result()

    # STEP 8 - Copy the result CSV from stage to archive, then delete from stage
    # Calculate the fpaths we will use ahead of time
    result_stg_fpath = os_usg_configs["results_stg_gcs_fpth"] % (
        datetime.strptime(args.date, "%Y-%m-%d").date() - timedelta(days=4),
        args.date,
    )
    result_archive_fpath = os_usg_configs["results_archive_gcs_fpth"] % (
        datetime.strptime(args.date, "%Y-%m-%d").date() - timedelta(days=4),
        args.date,
    )
    move_blob(
        os_usg_configs["bucket_no_gs"],
        result_stg_fpath,
        os_usg_configs["bucket_no_gs"],
        result_archive_fpath,
    )

    # STEP 9 - Copy the error CSV from stage to archive, then delete from stage
    error_stg_fpath = os_usg_configs["errors_stg_gcs_fpth"] % (
        datetime.strptime(args.date, "%Y-%m-%d").date() - timedelta(days=4),
        args.date,
    )
    error_archive_fpath = os_usg_configs["errors_archive_gcs_fpth"] % (
        datetime.strptime(args.date, "%Y-%m-%d").date() - timedelta(days=4),
        args.date,
    )
    move_blob(
        os_usg_configs["bucket_no_gs"],
        error_stg_fpath,
        os_usg_configs["bucket_no_gs"],
        error_archive_fpath,
    )


if __name__ == "__main__":
    main()



"""Telemetry probes data - download from API, clean and upload to BigQuery."""

import logging
from pathlib import Path

import click
import requests
import yaml
from google.cloud import bigquery

API_BASE_URL = "https://probeinfo.telemetry.mozilla.org"

DEFAULT_PROJECT_ID = Path(__file__).parent.parent.parent.name
DEFAULT_DATASET_ID = Path(__file__).parent.parent.name
DEFAULT_TABLE_NAME = Path(__file__).parent.name
CHANNELS = ["release", "beta", "nightly"]

SCHEMA_FILE = Path(__file__).parent / "schema.yaml"
SCHEMA = bigquery.SchemaField.from_api_repr(
    {"name": "root", "type": "RECORD", **yaml.safe_load(SCHEMA_FILE.read_text())}
).fields


def get_api_response(url):
    """Get json of response if the requests to the API was successful."""
    response = requests.get(url)
    response.raise_for_status()
    return response.json()


def store_data_in_bigquery(data, schema, destination_project, destination_table_id):
    """Upload data to Bigquery in a single, non partitioned table."""
    client = bigquery.Client(project=destination_project)

    job_config = bigquery.LoadJobConfig(
        create_disposition="CREATE_IF_NEEDED",
        schema=schema,
        source_format=bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
        write_disposition="WRITE_TRUNCATE",
    )

    load_job = client.load_table_from_json(
        data, destination_table_id, location="US", job_config=job_config
    )
    load_job.result()
    stored_table = client.get_table(destination_table_id)
    logging.info(f"Loaded {stored_table.num_rows} rows into {destination_table_id}.")


def download_telemetry_probes(url: str):
    """Download probes for telemetry and parse the data."""
    firefox_metrics = []
    for channel in CHANNELS:
        probes = get_api_response(f"{url}/firefox/{channel}/main/all_probes")
        for probe in probes.values():
            release_version = int(probe["history"][channel][-1]["versions"]["first"])
            last_version = int(probe["history"][channel][0]["versions"]["last"])
            expiry_version = probe["history"][channel][0]["expiry_version"]
            first_added_date = probe["first_added"][channel][:10]

            firefox_metrics.append(
                {
                    "channel": channel,
                    "probe": probe["name"],
                    "type": probe["type"],
                    "release_version": release_version,
                    "last_version": last_version,
                    "expiry_version": expiry_version,
                    "first_added_date": first_added_date,
                }
            )
    return firefox_metrics


@click.command
@click.option(
    "--bq_project_id",
    default=DEFAULT_PROJECT_ID,
    show_default=True,
    help="BigQuery project the data is written to.",
)
@click.option(
    "--bq_dataset_id",
    default=DEFAULT_DATASET_ID,
    show_default=True,
    help="BigQuery dataset the data is written to.",
)
@click.option(
    "--bq_table_name",
    default=DEFAULT_TABLE_NAME,
    show_default=True,
    help="Bigquery table the data is written to.",
)
def run_telemetry_probes(bq_project_id, bq_dataset_id, bq_table_name):
    """Download the data for telemetry probes from the API ad store it in BigQuery."""
    telemetry_probes = download_telemetry_probes(API_BASE_URL)
    destination_table_id = f"{bq_project_id}.{bq_dataset_id}.{bq_table_name}"
    store_data_in_bigquery(
        data=telemetry_probes,
        schema=SCHEMA,
        destination_project=bq_project_id,
        destination_table_id=destination_table_id,
    )


if __name__ == "__main__":
    logging.getLogger().setLevel(logging.INFO)
    run_telemetry_probes()



"""Tool for detecting and deleting broken views."""

import logging
from functools import partial
from multiprocessing.pool import ThreadPool

from google.api_core.exceptions import Forbidden, NotFound
from google.cloud import bigquery

from bigquery_etl.cli.utils import table_matches_patterns
from bigquery_etl.util.bigquery_id import sql_table_id


def list_broken_views(project_id, parallelism, only, log_level):
    """Dry run all views in the configured project and report on broken ones."""
    try:
        logging.basicConfig(level=log_level, format="%(levelname)s %(message)s")
    except ValueError as e:
        print(f"argument --log-level: {e}")

    client = bigquery.Client(project_id)

    table_filter = partial(table_matches_patterns, "*", False)

    if only:
        table_filter = partial(table_matches_patterns, only, False)

    with ThreadPool(parallelism) as pool:
        views = list_views(
            client=client,
            pool=pool,
            project_id=project_id,
            only_tables=[only] if only is not None else [],
            table_filter=table_filter,
        )
        print(views)
        results = pool.map(partial(dry_run_view, client), views)
    print("Failed with Forbidden error:")
    for result in results:
        status, view = result
        if status == "Forbidden":
            print(f"  {view}")
    print("\nFailed with NotFound error; run the following commands to remove:")
    for result in results:
        status, view = result
        if status == "NotFound":
            print(f"  bq rm -f {view.replace('.', ':', 1)}")


def list_views(client, pool, project_id, only_tables, table_filter):
    """Make parallel BQ API calls to grab all views.

    See `util.standard_ for more context on table filtering.

    :param client:       A BigQuery client object
    :param pool:         A process pool for handling concurrent calls
    :param project_id:   Target project
    :param only_tables:  An iterable of globs in `<dataset>.<view>` format
    :param table_filter: A function for determining whether to include a view
    :return:             A list of matching views
    """
    if only_tables and not _contains_glob(only_tables):
        # skip list calls when only_tables exists and contains no globs
        return [f"{project_id}.{t}" for t in only_tables if table_filter(t)]
    if only_tables and not _contains_glob(
        _extract_dataset_from_glob(t) for t in only_tables
    ):
        # skip list_datasets call when only_tables exists and datasets contain no globs
        datasets = {
            f"{project_id}.{_extract_dataset_from_glob(t)}" for t in only_tables
        }
    else:
        datasets = [d.reference.dataset_id for d in client.list_datasets(project_id)]

    return [
        sql_table_id(t)
        for tables in pool.map(client.list_tables, datasets)
        for t in tables
        if table_filter(f"{t.dataset_id}.{t.table_id}") and t.table_type == "VIEW"
    ]


def dry_run_view(client, view):
    """Dry run a view, returning a tuple of (status, view_name)."""
    sql = (
        "CREATE OR REPLACE VIEW telemetry.dry_run_view AS "
        + client.get_table(view).view_query
    )
    job_config = bigquery.QueryJobConfig(dry_run=True, use_legacy_sql=False)
    try:
        client.query(sql, job_config)
    except NotFound:
        return ("NotFound", view)
    except Forbidden:
        return ("Forbidden", view)
    else:
        return ("Success", view)


def _contains_glob(patterns):
    return any({"*", "?", "["}.intersection(pattern) for pattern in patterns)


def _extract_dataset_from_glob(pattern):
    # Assumes globs are in <dataset>.<table> form without a project specified.
    return pattern.split(".", 1)[0]



"""
Generate app (as opposed to channel) specific views for Glean ping tables.

The generation logic sets fields that might not be present in stable tables
(but are present in others) to NULL. Fields are ordered so that UNIONs across
the stable tables are possible.
For views that have incomaptible schemas (e.g due to fields having mismatching
types), the view is only generated for the release channel.
"""

import os
from copy import deepcopy
from pathlib import Path

from jinja2 import Environment, FileSystemLoader
from mozilla_schema_generator.glean_ping import GleanPing
from pathos.multiprocessing import ThreadingPool

from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.schema import Schema
from bigquery_etl.util.common import get_table_dir, write_sql
from sql_generators.glean_usage.common import GleanTable

VIEW_METADATA_TEMPLATE = """\
# Generated by bigquery_etl.glean_usage.GleanAppPingViews
---
friendly_name: App-specific view for Glean ping "{ping_name}"
description: |-
  This a view that UNIONs the stable ping tables
  across all channels of the Glean application "{app_name}"
  ({app_channels}).

  It is used by Looker.
"""

# Fields that exist in the source dataset,
# but are manually overriden in the constructed SQL.
# MUST be kept in sync with the query in `app_ping_view.view.sql`
OVERRIDDEN_FIELDS = {"normalized_channel"}

VIEW_SQL_LENGTH_LIMIT = 1024 * 1024

PATH = Path(os.path.dirname(__file__))


class GleanAppPingViews(GleanTable):
    """Represents generated Glean app ping view."""

    def __init__(self):
        """Initialize Glean ping view."""
        GleanTable.__init__(self)
        self.per_app_id_enabled = False
        self.per_app_enabled = True

    def generate_per_app(
        self,
        project_id,
        app_info,
        output_dir=None,
        use_cloud_function=True,
        parallelism=8,
        id_token=None,
    ):
        """
        Generate per-app ping views across channels.

        If schemas are incompatible, then use release channel only.
        """
        # get release channel info
        release_app = app_info[0]
        target_dataset = release_app["app_name"]

        # channels are all in the same repo, sending the same pings
        repo = next(
            (r for r in GleanPing.get_repos() if r["name"] == release_app["v1_name"])
        )

        # app name is the same as the bq_dataset_family for the release channel: do nothing
        if (
            repo["app_id"] == release_app["app_name"]
            or release_app["bq_dataset_family"] == release_app["app_name"]
        ):
            return

        env = Environment(loader=FileSystemLoader(PATH / "templates"))
        view_template = env.get_template("app_ping_view.view.sql")

        skip_existing = self.skip_existing(output_dir, project_id)

        p = GleanPing(repo)
        # generate views for all available pings

        def _process_ping(ping_name):
            view_name = ping_name.replace("-", "_")
            full_view_id = f"moz-fx-data-shared-prod.{target_dataset}.{view_name}"

            # generate a unioned schema that contains all fields of all ping tables across all channels
            unioned_schema = Schema.empty()

            # cache schemas to be reused when generating the select expression
            cached_schemas = {}

            # iterate through app_info to get all channels
            included_channel_apps = []
            included_channel_views = []
            for channel_app in app_info:
                channel_dataset = channel_app["bq_dataset_family"]
                channel_dataset_view = f"{channel_dataset}.{view_name}"
                schema = Schema.for_table(
                    "moz-fx-data-shared-prod",
                    channel_dataset,
                    view_name,
                    partitioned_by="submission_timestamp",
                    use_cloud_function=use_cloud_function,
                    id_token=id_token,
                )
                cached_schemas[channel_dataset] = deepcopy(schema)

                if schema.schema["fields"] == []:
                    # check for empty schemas (e.g. restricted ones) and skip for now
                    print(f"Cannot get schema for `{channel_dataset_view}`; Skipping")
                    continue

                try:
                    unioned_schema.merge(
                        schema, add_missing_fields=True, ignore_incompatible_fields=True
                    )
                except Exception as e:
                    # if schema incompatibilities are detected, then only generate for release channel
                    print(
                        f"Cannot UNION `moz-fx-data-shared-prod.{channel_dataset_view}`: {e}"
                    )
                    break

                included_channel_apps.append(channel_app)
                included_channel_views.append(channel_dataset_view)

            if included_channel_apps == []:
                # nothing to render
                return

            # generate the SELECT expression used for UNIONing the stable tables;
            # fields that are not part of a table, but exist in others, are set to NULL
            def _generate_view_sql(restructure_metrics=False) -> str:
                queries = []
                for channel_app in included_channel_apps:
                    channel_dataset = channel_app["bq_dataset_family"]

                    # compare table schema with unioned schema to determine fields that need to be NULL
                    select_expression = cached_schemas[
                        channel_dataset
                    ].generate_compatible_select_expression(
                        unioned_schema,
                        fields_to_remove=OVERRIDDEN_FIELDS,
                        unnest_structs=restructure_metrics,
                        max_unnest_depth=2,
                        unnest_allowlist="metrics",
                    )

                    queries.append(
                        dict(
                            select_expression=select_expression,
                            dataset=channel_dataset,
                            table=view_name,
                            channel=channel_app.get("app_channel"),
                            app_name=release_app["app_name"],
                            includes_client_info=any(
                                [
                                    "client_info" == f["name"]
                                    for f in unioned_schema.schema["fields"]
                                ]
                            ),
                        )
                    )

                render_kwargs = dict(
                    project_id=project_id, target_view=full_view_id, queries=queries
                )
                return reformat(view_template.render(**render_kwargs))

            view_sql = _generate_view_sql(restructure_metrics=True)
            if len(view_sql) > VIEW_SQL_LENGTH_LIMIT:
                print(
                    f"Generated SQL for `{full_view_id}` view with restructured `metrics` exceeds {VIEW_SQL_LENGTH_LIMIT:,} character limit."
                    " Regenerating SQL without restructured `metrics`."
                )
                view_sql = _generate_view_sql(restructure_metrics=False)

            # write generated SQL files to destination folders
            if output_dir:
                write_sql(
                    output_dir,
                    full_view_id,
                    "view.sql",
                    view_sql,
                    skip_existing=str(
                        get_table_dir(output_dir, full_view_id) / "view.sql"
                    )
                    in skip_existing,
                )

                write_sql(
                    output_dir,
                    full_view_id,
                    "metadata.yaml",
                    VIEW_METADATA_TEMPLATE.format(
                        ping_name=ping_name,
                        app_name=release_app["canonical_app_name"],
                        app_channels=", ".join(included_channel_views),
                    ),
                    skip_existing=str(
                        get_table_dir(output_dir, full_view_id) / "metadata.yaml"
                    )
                    in skip_existing,
                )

                schema_dir = get_table_dir(output_dir, full_view_id)

                # remove overridden fields from schema
                # it's assumed that these fields are added separately, or ignored completely
                unioned_schema.schema["fields"] = [
                    field
                    for field in unioned_schema.schema["fields"]
                    if field["name"] not in OVERRIDDEN_FIELDS
                ]

                # normalized_app_id is not part of the underlying table the schemas are derived from,
                # the field gets added as part of the view definition, so we have to add it manually to the schema
                unioned_schema.schema["fields"] = [
                    {
                        "name": "normalized_app_id",
                        "mode": "NULLABLE",
                        "type": "STRING",
                        "description": "App ID of the channel data was received from",
                    },
                    {
                        "name": "normalized_channel",
                        "mode": "NULLABLE",
                        "type": "STRING",
                        "description": "Normalized channel name",
                    },
                ] + unioned_schema.schema["fields"]

                unioned_schema.to_yaml_file(schema_dir / "schema.yaml")

        # Using ThreadingPool instead of ProcessingPool here, due to issues with pickling the GleanAppPingViews class (self references),
        # and ProcessingPools cannot be nested - glean_usage is using ProcessingPool to kick off generating the different tables per app
        with ThreadingPool(parallelism) as pool:
            pool.map(
                _process_ping,
                p.get_pings(),
            )



"""Testing data for query."""

from pathlib import Path

import yaml

ROOT = Path(__file__).parent

UUID = "df735f02-efe5-4b07-b212-583bb99ba241"
SUBMISSION_DATE = "2020-10-01"
APP_BUILD_ID = "2020100100"

# NOTE: what happens when channel = "*"?
HISTOGRAM_BUCKET_COUNTS = [
    {
        "agg_type": "summed_histogram",
        "app_build_id": "*",
        "app_version": 84,
        "channel": "*",
        "key": "",
        "metric": "network_tcp_connection",
        "metric_type": "timing_distribution",
        "os": "*",
        "ping_type": "*",
        "range_max": 3,
        "record": {"key": "1", "value": 1.0},
    },
    {
        "agg_type": "summed_histogram",
        "app_build_id": "*",
        "app_version": 84,
        "channel": "*",
        "key": "",
        "metric": "network_tcp_connection",
        "metric_type": "timing_distribution",
        "os": "*",
        "ping_type": "*",
        "range_max": 3,
        "record": {"key": "2", "value": 0.0},
    },
]

EXPECT = [
    {
        "agg_type": "histogram",
        "aggregates": [
            {"key": "0", "value": 0.125},
            {"key": "1", "value": 0.625},
            {"key": "2", "value": 0.125},
            {"key": "3", "value": 0.125},
        ],
        "app_build_id": "*",
        "app_version": 84,
        "channel": "*",
        "client_agg_type": "summed_histogram",
        "key": "",
        "metric": "network_tcp_connection",
        "metric_type": "timing_distribution",
        "os": "*",
        "ping_type": "*",
        "total_users": 1,
    }
]

prefix = "glam_etl"
tables = [
    (
        f"{prefix}.org_mozilla_fenix_glam_nightly__histogram_bucket_counts_v1.yaml",
        HISTOGRAM_BUCKET_COUNTS,
    ),
    ("expect.yaml", EXPECT),
]
for name, data in tables:
    with (ROOT / name).open("w") as fp:
        yaml.dump(data, fp)
